/* autores_bucket1d.c : 1-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_bucket1d.h"



/* Bucket1D */

void AutoRes_Bucket1D_initialize
(AutoRes_Bucket1D *self,
 double minRangeX, 
 double maxRangeX,
 int xCells, 
 int nAverageItemsPerCell)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx;

  assert(minRangeX + tolerance < maxRangeX);
  assert(0 < xCells);
  assert(0 < nAverageItemsPerCell);

  self->minRangeX = minRangeX;
  self->maxRangeX = maxRangeX;
  self->xCells = xCells;

  self->startLinkIndexs = (int*)malloc (sizeof (int) * xCells);
  assert(self->startLinkIndexs != NULL);

  for (cx = 0; cx < xCells; cx++) {
    self->startLinkIndexs[cx] = -1;
  }

  self->startLinkIndexOutOfRange = -1;
  self->startLinkIndexOfFreePool = -1;

  self->maxLinks = xCells * nAverageItemsPerCell;
  self->nLinks = 0;

  self->nextLinkIndexs = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->nextLinkIndexs != NULL);

  self->linkItemIds = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->linkItemIds != NULL);
}

void AutoRes_Bucket1D_terminate
(AutoRes_Bucket1D *self)
{
  free (self->startLinkIndexs);
  free (self->nextLinkIndexs);
  free (self->linkItemIds);
}

void AutoRes_Bucket1D_getMinRange 
(AutoRes_Bucket1D *self,
 double *minRangeX_OUT)
{
  *minRangeX_OUT = self->minRangeX;
}

void AutoRes_Bucket1D_getMaxRange 
(AutoRes_Bucket1D *self,
 double *maxRangeX_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
}

void AutoRes_Bucket1D_getNCells 
(AutoRes_Bucket1D *self,
 int *xCells_OUT)
{
  *xCells_OUT = self->xCells;
}

void AutoRes_Bucket1D_getIndexWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int *cx0_OUT, 
 int *cx1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX;
  double div0X;
  int cx0;
  double div1X;
  int cx1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  cx1 = (int)floor (div1X + tolerance);

  *cx0_OUT = cx0;
  *cx1_OUT = cx1;

  assert(*cx0_OUT <= *cx1_OUT);
}

void AutoRes_Bucket1D_getIndexAt 
(AutoRes_Bucket1D *self,
 double x,
 int *cx_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX;
  double divX;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;

  divX = (x - self->minRangeX) / cellSizeX;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
}

#ifdef NDEBUG
#else

int AutoRes_Bucket1D_isOutOfRange 
(AutoRes_Bucket1D *self,
 int cx)
{
  return cx < 0 || self->xCells <= cx;
}

int AutoRes_Bucket1D_startLinkIndexIn 
(AutoRes_Bucket1D *self,
 int cx)
{
  assert(0 <= cx && cx < self->xCells);

  return self->startLinkIndexs[cx];
}

int AutoRes_Bucket1D_startLinkIndexOutOfRange 
(AutoRes_Bucket1D *self)
{
  return self->startLinkIndexOutOfRange;
}

int AutoRes_Bucket1D_nextLinkIndex 
(AutoRes_Bucket1D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->nextLinkIndexs[linkIndex];
}

int AutoRes_Bucket1D_itemId 
(AutoRes_Bucket1D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->linkItemIds[linkIndex];
}

#endif /* NDEBUG */

static void AutoRes_Bucket1D_addLinkIntoFreePool 
(AutoRes_Bucket1D *self,
 int linkIndex)
{
  int startLinkIndex;

  assert(0 <= linkIndex && linkIndex < self->nLinks);
  
  startLinkIndex = self->startLinkIndexOfFreePool;
  self->startLinkIndexOfFreePool = linkIndex;
  self->nextLinkIndexs[linkIndex] = startLinkIndex;
  self->linkItemIds[linkIndex] = -1;
}

static int AutoRes_Bucket1D_removeLinkFromFreePool 
(AutoRes_Bucket1D *self)
{
  int linkIndex;

  linkIndex = self->startLinkIndexOfFreePool;
  if (linkIndex != -1) {
    self->startLinkIndexOfFreePool = self->nextLinkIndexs[linkIndex];
  }

  assert(linkIndex == -1 
	 || (0 <= linkIndex && linkIndex < self->nLinks));

  return linkIndex;
}

static int AutoRes_Bucket1D_makeNewLink 
(AutoRes_Bucket1D *self)
{
  int newLinkIndex;

  newLinkIndex = AutoRes_Bucket1D_removeLinkFromFreePool
    (self);
  if (newLinkIndex == -1) {
    assert(self->nLinks < self->maxLinks);
    
    newLinkIndex = self->nLinks;
    self->nLinks++;
  }

  return newLinkIndex;
}

static void AutoRes_Bucket1D_addItemIn 
(AutoRes_Bucket1D *self,
 int cx, int itemId)
{
  int listIndex;
  int newLinkIndex;
  int startLinkIndex;

  assert(0 <= cx && cx < self->xCells);
  
  listIndex = cx;

  newLinkIndex = AutoRes_Bucket1D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexs[listIndex];
  self->startLinkIndexs[listIndex] = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket1D_addItemOutOfRange 
(AutoRes_Bucket1D *self,
 int itemId)
{
  int newLinkIndex;
  int startLinkIndex;

  newLinkIndex = AutoRes_Bucket1D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexOutOfRange;
  self->startLinkIndexOutOfRange = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket1D_removeItemIn 
(AutoRes_Bucket1D *self,
 int cx, int itemId)
{
  int listIndex;
  int currentLinkIndex;
  int nextLinkIndex;

  assert(0 <= cx && cx < self->xCells);
  
  listIndex = cx;

  currentLinkIndex = self->startLinkIndexs[listIndex];
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexs[listIndex]
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket1D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket1D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

static void AutoRes_Bucket1D_removeItemOutOfRange 
(AutoRes_Bucket1D *self,
 int itemId)
{
  int currentLinkIndex;
  int nextLinkIndex;

  currentLinkIndex = self->startLinkIndexOutOfRange;
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexOutOfRange
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket1D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket1D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

void AutoRes_Bucket1D_addItemWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0;
  int cx1;
  int cx;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);

  AutoRes_Bucket1D_getIndexWithin 
    (self,
     minRangeX, 
     maxRangeX, 
     &cx0, 
     &cx1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    if (AutoRes_Bucket1D_isOutOfRange 
	(self,
	 cx)) {
      outOfRangeExists = 1;
    } else {
      AutoRes_Bucket1D_addItemIn 
	(self, 
	 cx, itemId);
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket1D_addItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket1D_addItemAt 
(AutoRes_Bucket1D *self,
 double x,
 int itemId)
{
  AutoRes_Bucket1D_addItemWithin 
    (self,
     x, 
     x, 
     itemId);
}

void AutoRes_Bucket1D_removeItemWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0;
  int cx1;
  int cx;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);

  AutoRes_Bucket1D_getIndexWithin 
    (self,
     minRangeX, 
     maxRangeX, 
     &cx0, 
     &cx1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    if (AutoRes_Bucket1D_isOutOfRange 
	(self,
	 cx)) {
      outOfRangeExists = 1;
    } else {
      AutoRes_Bucket1D_removeItemIn 
	(self, 
	 cx, itemId);
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket1D_removeItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket1D_removeItemAt 
(AutoRes_Bucket1D *self,
 double x,
 int itemId)
{
  AutoRes_Bucket1D_removeItemWithin 
    (self,
     x, 
     x, 
     itemId);
}



