/* autores_bucket2d.c : 2-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_bucket2d.h"



/* Bucket2D */

void AutoRes_Bucket2D_initialize
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY, 
 double maxRangeX, double maxRangeY,
 int xCells, int yCells, 
 int nAverageItemsPerCell)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx, cy;

  assert(minRangeX + tolerance < maxRangeX);
  assert(minRangeY + tolerance < maxRangeY);
  assert(0 < xCells);
  assert(0 < yCells);
  assert(0 < nAverageItemsPerCell);

  self->minRangeX = minRangeX;
  self->minRangeY = minRangeY;
  self->maxRangeX = maxRangeX;
  self->maxRangeY = maxRangeY;
  self->xCells = xCells;
  self->yCells = yCells;

  self->startLinkIndexs = (int*)malloc (sizeof (int) * xCells * yCells);
  assert(self->startLinkIndexs != NULL);

  for (cx = 0; cx < xCells; cx++) {
    for (cy = 0; cy < yCells; cy++) {
      self->startLinkIndexs[cx * yCells + cy] = -1;
    }
  }

  self->startLinkIndexOutOfRange = -1;
  self->startLinkIndexOfFreePool = -1;

  self->maxLinks = xCells * yCells * nAverageItemsPerCell;
  self->nLinks = 0;

  self->nextLinkIndexs = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->nextLinkIndexs != NULL);

  self->linkItemIds = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->linkItemIds != NULL);
}

void AutoRes_Bucket2D_terminate
(AutoRes_Bucket2D *self)
{
  free (self->startLinkIndexs);
  free (self->nextLinkIndexs);
  free (self->linkItemIds);
}

void AutoRes_Bucket2D_getMinRange 
(AutoRes_Bucket2D *self,
 double *minRangeX_OUT, double *minRangeY_OUT)
{
  *minRangeX_OUT = self->minRangeX;
  *minRangeY_OUT = self->minRangeY;
}

void AutoRes_Bucket2D_getMaxRange 
(AutoRes_Bucket2D *self,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
  *maxRangeY_OUT = self->maxRangeY;
}

void AutoRes_Bucket2D_getNCells 
(AutoRes_Bucket2D *self,
 int *xCells_OUT, int *yCells_OUT)
{
  *xCells_OUT = self->xCells;
  *yCells_OUT = self->yCells;
}

void AutoRes_Bucket2D_getIndexWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int *cx0_OUT, int *cy0_OUT, 
 int *cx1_OUT, int *cy1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY;
  double div0X, div0Y;
  int cx0, cy0;
  double div1X, div1Y;
  int cx1, cy1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  div0Y = (minRangeY - self->minRangeY) / cellSizeY;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;
  cy0 = (int)floor (div0Y + tolerance);
  if (fabs (div0Y - cy0) < tolerance) cy0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  div1Y = (maxRangeY - self->minRangeY) / cellSizeY;
  cx1 = (int)floor (div1X + tolerance);
  cy1 = (int)floor (div1Y + tolerance);

  *cx0_OUT = cx0;
  *cy0_OUT = cy0;
  *cx1_OUT = cx1;
  *cy1_OUT = cy1;

  assert(*cx0_OUT <= *cx1_OUT);
  assert(*cy0_OUT <= *cy1_OUT);
}

void AutoRes_Bucket2D_getIndexAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int *cx_OUT, int *cy_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY;
  double divX, divY;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;

  divX = (x - self->minRangeX) / cellSizeX;
  divY = (y - self->minRangeY) / cellSizeY;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
  *cy_OUT = (int)floor (divY + tolerance);
  if (fabs (y - self->maxRangeY) < tolerance) *cy_OUT = self->yCells - 1;
}

#ifdef NDEBUG
#else

int AutoRes_Bucket2D_isOutOfRange 
(AutoRes_Bucket2D *self,
 int cx, int cy)
{
  return cx < 0 || self->xCells <= cx 
    || cy < 0 || self->yCells <= cy;
}

int AutoRes_Bucket2D_startLinkIndexIn 
(AutoRes_Bucket2D *self,
 int cx, int cy)
{
  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells);

  return self->startLinkIndexs[cx * self->yCells + cy];
}

int AutoRes_Bucket2D_startLinkIndexOutOfRange 
(AutoRes_Bucket2D *self)
{
  return self->startLinkIndexOutOfRange;
}

int AutoRes_Bucket2D_nextLinkIndex 
(AutoRes_Bucket2D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->nextLinkIndexs[linkIndex];
}

int AutoRes_Bucket2D_itemId 
(AutoRes_Bucket2D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->linkItemIds[linkIndex];
}

#endif /* NDEBUG */

static void AutoRes_Bucket2D_addLinkIntoFreePool 
(AutoRes_Bucket2D *self,
 int linkIndex)
{
  int startLinkIndex;

  assert(0 <= linkIndex && linkIndex < self->nLinks);
  
  startLinkIndex = self->startLinkIndexOfFreePool;
  self->startLinkIndexOfFreePool = linkIndex;
  self->nextLinkIndexs[linkIndex] = startLinkIndex;
  self->linkItemIds[linkIndex] = -1;
}

static int AutoRes_Bucket2D_removeLinkFromFreePool 
(AutoRes_Bucket2D *self)
{
  int linkIndex;

  linkIndex = self->startLinkIndexOfFreePool;
  if (linkIndex != -1) {
    self->startLinkIndexOfFreePool = self->nextLinkIndexs[linkIndex];
  }

  assert(linkIndex == -1 
	 || (0 <= linkIndex && linkIndex < self->nLinks));

  return linkIndex;
}

static int AutoRes_Bucket2D_makeNewLink 
(AutoRes_Bucket2D *self)
{
  int newLinkIndex;

  newLinkIndex = AutoRes_Bucket2D_removeLinkFromFreePool
    (self);
  if (newLinkIndex == -1) {
    assert(self->nLinks < self->maxLinks);
    
    newLinkIndex = self->nLinks;
    self->nLinks++;
  }

  return newLinkIndex;
}

static void AutoRes_Bucket2D_addItemIn 
(AutoRes_Bucket2D *self,
 int cx, int cy, int itemId)
{
  int listIndex;
  int newLinkIndex;
  int startLinkIndex;

  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells);
  
  listIndex = cx * self->yCells + cy;

  newLinkIndex = AutoRes_Bucket2D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexs[listIndex];
  self->startLinkIndexs[listIndex] = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket2D_addItemOutOfRange 
(AutoRes_Bucket2D *self,
 int itemId)
{
  int newLinkIndex;
  int startLinkIndex;

  newLinkIndex = AutoRes_Bucket2D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexOutOfRange;
  self->startLinkIndexOutOfRange = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket2D_removeItemIn 
(AutoRes_Bucket2D *self,
 int cx, int cy, int itemId)
{
  int listIndex;
  int currentLinkIndex;
  int nextLinkIndex;

  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells);
  
  listIndex = cx * self->yCells + cy;

  currentLinkIndex = self->startLinkIndexs[listIndex];
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexs[listIndex]
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket2D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket2D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

static void AutoRes_Bucket2D_removeItemOutOfRange 
(AutoRes_Bucket2D *self,
 int itemId)
{
  int currentLinkIndex;
  int nextLinkIndex;

  currentLinkIndex = self->startLinkIndexOutOfRange;
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexOutOfRange
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket2D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket2D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

void AutoRes_Bucket2D_addItemWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0;
  int cx1, cy1;
  int cx, cy;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);

  AutoRes_Bucket2D_getIndexWithin 
    (self,
     minRangeX, minRangeY, 
     maxRangeX, maxRangeY, 
     &cx0, &cy0, 
     &cx1, &cy1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      if (AutoRes_Bucket2D_isOutOfRange 
	  (self,
	   cx, cy)) {
	outOfRangeExists = 1;
      } else {
	AutoRes_Bucket2D_addItemIn 
	  (self, 
	   cx, cy, itemId);
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket2D_addItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket2D_addItemAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int itemId)
{
  AutoRes_Bucket2D_addItemWithin 
    (self,
     x, y, 
     x, y, 
     itemId);
}

void AutoRes_Bucket2D_removeItemWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0;
  int cx1, cy1;
  int cx, cy;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);

  AutoRes_Bucket2D_getIndexWithin 
    (self,
     minRangeX, minRangeY, 
     maxRangeX, maxRangeY, 
     &cx0, &cy0, 
     &cx1, &cy1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      if (AutoRes_Bucket2D_isOutOfRange 
	  (self,
	   cx, cy)) {
	outOfRangeExists = 1;
      } else {
	AutoRes_Bucket2D_removeItemIn 
	  (self, 
	   cx, cy, itemId);
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket2D_removeItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket2D_removeItemAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int itemId)
{
  AutoRes_Bucket2D_removeItemWithin 
    (self,
     x, y, 
     x, y, 
     itemId);
}



