/* autores_bucket3d.c : 3-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_bucket3d.h"



/* Bucket3D */

void AutoRes_Bucket3D_initialize
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ, 
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int xCells, int yCells, int zCells, 
 int nAverageItemsPerCell)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx, cy, cz;

  assert(minRangeX + tolerance < maxRangeX);
  assert(minRangeY + tolerance < maxRangeY);
  assert(minRangeZ + tolerance < maxRangeZ);
  assert(0 < xCells);
  assert(0 < yCells);
  assert(0 < zCells);
  assert(0 < nAverageItemsPerCell);

  self->minRangeX = minRangeX;
  self->minRangeY = minRangeY;
  self->minRangeZ = minRangeZ;
  self->maxRangeX = maxRangeX;
  self->maxRangeY = maxRangeY;
  self->maxRangeZ = maxRangeZ;
  self->xCells = xCells;
  self->yCells = yCells;
  self->zCells = zCells;

  self->startLinkIndexs 
    = (int*)malloc (sizeof (int) * xCells * yCells * zCells);
  assert(self->startLinkIndexs != NULL);

  for (cx = 0; cx < xCells; cx++) {
    for (cy = 0; cy < yCells; cy++) {
      for (cz = 0; cz < zCells; cz++) {
	self->startLinkIndexs
	  [cx * yCells * zCells + cy * zCells + cz] = -1;
      }
    }
  }

  self->startLinkIndexOutOfRange = -1;
  self->startLinkIndexOfFreePool = -1;

  self->maxLinks = xCells * yCells * zCells * nAverageItemsPerCell;
  self->nLinks = 0;



  /* NEED TO REDEFINE */
  fprintf (stderr, 
	   " xCells %d yCells %d zCells %d items %d maxLinks %d \n",
	   xCells, yCells, zCells, nAverageItemsPerCell,
	   self->maxLinks);




  self->nextLinkIndexs = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->nextLinkIndexs != NULL);

  self->linkItemIds = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->linkItemIds != NULL);
}

void AutoRes_Bucket3D_terminate
(AutoRes_Bucket3D *self)
{
  free (self->startLinkIndexs);
  free (self->nextLinkIndexs);
  free (self->linkItemIds);
}

void AutoRes_Bucket3D_getMinRange 
(AutoRes_Bucket3D *self,
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT)
{
  *minRangeX_OUT = self->minRangeX;
  *minRangeY_OUT = self->minRangeY;
  *minRangeZ_OUT = self->minRangeZ;
}

void AutoRes_Bucket3D_getMaxRange 
(AutoRes_Bucket3D *self,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
  *maxRangeY_OUT = self->maxRangeY;
  *maxRangeZ_OUT = self->maxRangeZ;
}

void AutoRes_Bucket3D_getNCells 
(AutoRes_Bucket3D *self,
 int *xCells_OUT, int *yCells_OUT, int *zCells_OUT)
{
  *xCells_OUT = self->xCells;
  *yCells_OUT = self->yCells;
  *zCells_OUT = self->zCells;
}

void AutoRes_Bucket3D_getIndexWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int *cx0_OUT, int *cy0_OUT, int *cz0_OUT, 
 int *cx1_OUT, int *cy1_OUT, int *cz1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY, cellSizeZ;
  double div0X, div0Y, div0Z;
  int cx0, cy0, cz0;
  double div1X, div1Y, div1Z;
  int cx1, cy1, cz1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;
  cellSizeZ = (self->maxRangeZ - self->minRangeZ) / self->zCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  div0Y = (minRangeY - self->minRangeY) / cellSizeY;
  div0Z = (minRangeZ - self->minRangeZ) / cellSizeZ;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;
  cy0 = (int)floor (div0Y + tolerance);
  if (fabs (div0Y - cy0) < tolerance) cy0--;
  cz0 = (int)floor (div0Z + tolerance);
  if (fabs (div0Z - cz0) < tolerance) cz0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  div1Y = (maxRangeY - self->minRangeY) / cellSizeY;
  div1Z = (maxRangeZ - self->minRangeZ) / cellSizeZ;
  cx1 = (int)floor (div1X + tolerance);
  cy1 = (int)floor (div1Y + tolerance);
  cz1 = (int)floor (div1Z + tolerance);

  *cx0_OUT = cx0;
  *cy0_OUT = cy0;
  *cz0_OUT = cz0;
  *cx1_OUT = cx1;
  *cy1_OUT = cy1;
  *cz1_OUT = cz1;

  assert(*cx0_OUT <= *cx1_OUT);
  assert(*cy0_OUT <= *cy1_OUT);
  assert(*cz0_OUT <= *cz1_OUT);
}

void AutoRes_Bucket3D_getIndexAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int *cx_OUT, int *cy_OUT, int *cz_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY, cellSizeZ;
  double divX, divY, divZ;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;
  cellSizeZ = (self->maxRangeZ - self->minRangeZ) / self->zCells;

  divX = (x - self->minRangeX) / cellSizeX;
  divY = (y - self->minRangeY) / cellSizeY;
  divZ = (z - self->minRangeZ) / cellSizeZ;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
  *cy_OUT = (int)floor (divY + tolerance);
  if (fabs (y - self->maxRangeY) < tolerance) *cy_OUT = self->yCells - 1;
  *cz_OUT = (int)floor (divZ + tolerance);
  if (fabs (z - self->maxRangeZ) < tolerance) *cz_OUT = self->zCells - 1;
}

#ifdef NDEBUG
#else

int AutoRes_Bucket3D_isOutOfRange 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz)
{
  return cx < 0 || self->xCells <= cx 
    || cy < 0 || self->yCells <= cy
    || cz < 0 || self->zCells <= cz;
}

int AutoRes_Bucket3D_startLinkIndexIn 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz)
{
  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells
	 && 0 <= cz && cz < self->zCells);

  return self->startLinkIndexs
    [cx * self->yCells * self->zCells 
     + cy * self->zCells 
     + cz];
}

int AutoRes_Bucket3D_startLinkIndexOutOfRange 
(AutoRes_Bucket3D *self)
{
  return self->startLinkIndexOutOfRange;
}

int AutoRes_Bucket3D_nextLinkIndex 
(AutoRes_Bucket3D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->nextLinkIndexs[linkIndex];
}

int AutoRes_Bucket3D_itemId 
(AutoRes_Bucket3D *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->linkItemIds[linkIndex];
}

#endif /* NDEBUG */

static void AutoRes_Bucket3D_addLinkIntoFreePool 
(AutoRes_Bucket3D *self,
 int linkIndex)
{
  int startLinkIndex;

  assert(0 <= linkIndex && linkIndex < self->nLinks);
  
  startLinkIndex = self->startLinkIndexOfFreePool;
  self->startLinkIndexOfFreePool = linkIndex;
  self->nextLinkIndexs[linkIndex] = startLinkIndex;
  self->linkItemIds[linkIndex] = -1;
}

static int AutoRes_Bucket3D_removeLinkFromFreePool 
(AutoRes_Bucket3D *self)
{
  int linkIndex;

  linkIndex = self->startLinkIndexOfFreePool;
  if (linkIndex != -1) {
    self->startLinkIndexOfFreePool = self->nextLinkIndexs[linkIndex];
  }

  assert(linkIndex == -1 
	 || (0 <= linkIndex && linkIndex < self->nLinks));

  return linkIndex;
}

static int AutoRes_Bucket3D_makeNewLink 
(AutoRes_Bucket3D *self)
{
  int newLinkIndex;

  newLinkIndex = AutoRes_Bucket3D_removeLinkFromFreePool
    (self);
  if (newLinkIndex == -1) {
    assert(self->nLinks < self->maxLinks);
    
    newLinkIndex = self->nLinks;
    self->nLinks++;
  }

  return newLinkIndex;
}

static void AutoRes_Bucket3D_addItemIn 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz, int itemId)
{
  int listIndex;
  int newLinkIndex;
  int startLinkIndex;

  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells
	 && 0 <= cz && cz < self->zCells);
  
  listIndex = cx * self->yCells * self->zCells + cy * self->zCells + cz;

  newLinkIndex = AutoRes_Bucket3D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexs[listIndex];
  self->startLinkIndexs[listIndex] = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket3D_addItemOutOfRange 
(AutoRes_Bucket3D *self,
 int itemId)
{
  int newLinkIndex;
  int startLinkIndex;

  newLinkIndex = AutoRes_Bucket3D_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexOutOfRange;
  self->startLinkIndexOutOfRange = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

static void AutoRes_Bucket3D_removeItemIn 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz, int itemId)
{
  int listIndex;
  int currentLinkIndex;
  int nextLinkIndex;

  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells
	 && 0 <= cz && cz < self->zCells);
  
  listIndex = cx * self->yCells * self->zCells + cy * self->zCells + cz;

  currentLinkIndex = self->startLinkIndexs[listIndex];
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexs[listIndex]
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket3D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket3D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

static void AutoRes_Bucket3D_removeItemOutOfRange 
(AutoRes_Bucket3D *self,
 int itemId)
{
  int currentLinkIndex;
  int nextLinkIndex;

  currentLinkIndex = self->startLinkIndexOutOfRange;
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexOutOfRange
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_Bucket3D_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_Bucket3D_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

void AutoRes_Bucket3D_addItemWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0, cz0;
  int cx1, cy1, cz1;
  int cx, cy, cz;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);
  assert(minRangeZ <= maxRangeZ + tolerance);

  AutoRes_Bucket3D_getIndexWithin 
    (self,
     minRangeX, minRangeY, minRangeZ, 
     maxRangeX, maxRangeY, maxRangeZ, 
     &cx0, &cy0, &cz0, 
     &cx1, &cy1, &cz1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      for (cz = cz0; cz <= cz1; cz++) {
	if (AutoRes_Bucket3D_isOutOfRange 
	    (self,
	     cx, cy, cz)) {
	  outOfRangeExists = 1;
	} else {
	  AutoRes_Bucket3D_addItemIn 
	    (self, 
	     cx, cy, cz, itemId);
	}
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket3D_addItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket3D_addItemAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int itemId)
{
  AutoRes_Bucket3D_addItemWithin 
    (self,
     x, y, z, 
     x, y, z, 
     itemId);
}

void AutoRes_Bucket3D_removeItemWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int itemId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0, cz0;
  int cx1, cy1, cz1;
  int cx, cy, cz;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);
  assert(minRangeZ <= maxRangeZ + tolerance);

  AutoRes_Bucket3D_getIndexWithin 
    (self,
     minRangeX, minRangeY, minRangeZ, 
     maxRangeX, maxRangeY, maxRangeZ, 
     &cx0, &cy0, &cz0, 
     &cx1, &cy1, &cz1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      for (cz = cz0; cz <= cz1; cz++) {
	if (AutoRes_Bucket3D_isOutOfRange 
	    (self,
	     cx, cy, cz)) {
	  outOfRangeExists = 1;
	} else {
	  AutoRes_Bucket3D_removeItemIn 
	    (self, 
	     cx, cy, cz, itemId);
	}
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_Bucket3D_removeItemOutOfRange 
      (self, 
       itemId);
  }
}

void AutoRes_Bucket3D_removeItemAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int itemId)
{
  AutoRes_Bucket3D_removeItemWithin 
    (self,
     x, y, z, 
     x, y, z, 
     itemId);
}



