/* autores_listtable.c : table of multiple lists */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_listtable.h"



/* ListTable */

void AutoRes_ListTable_initialize
(AutoRes_ListTable *self,
 int nLists, 
 int nAverageLinksPerList)
{
  int iList;

  assert(0 < nLists);
  assert(0 < nAverageLinksPerList);

  self->nLists = nLists;

  self->startLinkIndexs = (int*)malloc (sizeof (int) * nLists);
  assert(self->startLinkIndexs != NULL);

  for (iList = 0; iList < nLists; iList++) {
    self->startLinkIndexs[iList] = -1;
  }

  self->startLinkIndexOfFreePool = -1;

  self->maxLinks = nLists * nAverageLinksPerList;
  self->nLinks = 0;

  self->nextLinkIndexs = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->nextLinkIndexs != NULL);

  self->linkItemIds = (int*)malloc (sizeof (int) * self->maxLinks);
  assert(self->linkItemIds != NULL);
}

void AutoRes_ListTable_terminate
(AutoRes_ListTable *self)
{
  free (self->startLinkIndexs);
  free (self->nextLinkIndexs);
  free (self->linkItemIds);
}

int AutoRes_ListTable_nLists 
(AutoRes_ListTable *self)
{
  return self->nLists;
}

#ifdef NDEBUG
#else

int AutoRes_ListTable_startLinkIndexIn 
(AutoRes_ListTable *self,
 int listIndex)
{
  assert(0 <= listIndex && listIndex < self->nLists);

  return self->startLinkIndexs[listIndex];
}

int AutoRes_ListTable_nextLinkIndex 
(AutoRes_ListTable *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->nextLinkIndexs[linkIndex];
}

int AutoRes_ListTable_itemId 
(AutoRes_ListTable *self,
 int linkIndex)
{
  assert(0 <= linkIndex && linkIndex < self->nLinks);

  return self->linkItemIds[linkIndex];
}

#endif /* NDEBUG */

static void AutoRes_ListTable_addLinkIntoFreePool 
(AutoRes_ListTable *self,
 int linkIndex)
{
  int startLinkIndex;

  assert(0 <= linkIndex && linkIndex < self->nLinks);
  
  startLinkIndex = self->startLinkIndexOfFreePool;
  self->startLinkIndexOfFreePool = linkIndex;
  self->nextLinkIndexs[linkIndex] = startLinkIndex;
  self->linkItemIds[linkIndex] = -1;
}

static int AutoRes_ListTable_removeLinkFromFreePool 
(AutoRes_ListTable *self)
{
  int linkIndex;

  linkIndex = self->startLinkIndexOfFreePool;
  if (linkIndex != -1) {
    self->startLinkIndexOfFreePool = self->nextLinkIndexs[linkIndex];
  }

  assert(linkIndex == -1 
	 || (0 <= linkIndex && linkIndex < self->nLinks));

  return linkIndex;
}

static int AutoRes_ListTable_makeNewLink 
(AutoRes_ListTable *self)
{
  int newLinkIndex;

  newLinkIndex = AutoRes_ListTable_removeLinkFromFreePool
    (self);
  if (newLinkIndex == -1) {
    assert(self->nLinks < self->maxLinks);
    
    newLinkIndex = self->nLinks;
    self->nLinks++;
  }

  return newLinkIndex;
}

void AutoRes_ListTable_addItemIn 
(AutoRes_ListTable *self,
 int listIndex, int itemId)
{
  int newLinkIndex;
  int startLinkIndex;

  assert(0 <= listIndex && listIndex < self->nLists);
  
  newLinkIndex = AutoRes_ListTable_makeNewLink 
    (self);
  startLinkIndex = self->startLinkIndexs[listIndex];
  self->startLinkIndexs[listIndex] = newLinkIndex;
  self->nextLinkIndexs[newLinkIndex] = startLinkIndex;
  self->linkItemIds[newLinkIndex] = itemId;
}

void AutoRes_ListTable_removeItemIn 
(AutoRes_ListTable *self,
 int listIndex, int itemId)
{
  int currentLinkIndex;
  int nextLinkIndex;

  assert(0 <= listIndex && listIndex < self->nLists);
  
  currentLinkIndex = self->startLinkIndexs[listIndex];
  if (currentLinkIndex == -1) return;

  assert(currentLinkIndex != -1);
  if (self->linkItemIds[currentLinkIndex] == itemId) {
    self->startLinkIndexs[listIndex]
      = self->nextLinkIndexs[currentLinkIndex];
    AutoRes_ListTable_addLinkIntoFreePool 
      (self, 
       currentLinkIndex);
    return;
  }

  for (;;) {
    assert(self->linkItemIds[currentLinkIndex] != itemId);

    if (self->nextLinkIndexs[currentLinkIndex] == -1) return;
    
    nextLinkIndex = self->nextLinkIndexs[currentLinkIndex];
    assert(nextLinkIndex != -1);

    if (self->linkItemIds[nextLinkIndex] == itemId) {
      self->nextLinkIndexs[currentLinkIndex]
	= self->nextLinkIndexs[nextLinkIndex];
      AutoRes_ListTable_addLinkIntoFreePool 
	(self, 
	 nextLinkIndex);
      return;
    }
    
    assert(self->linkItemIds[nextLinkIndex] != itemId);
    currentLinkIndex = nextLinkIndex;
  }
}

