/* autores_utility2d.c : 2-dimensional triangle mesh */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_listtable.h"
#include "autores_bucket2d.h"
#include "autores_utility2d.h"



/* Nodal Coordinate */

#define N_DIMENSIONS 2
#define X 0
#define Y 1
static int NNodes;
static double *NodalCoordinates;   /* [NNodes][N_DIMENSIONS] */



void AutoRes_SetNodalCoordinate2D 
(int nNodes,
 double nodalCoordinates[])
{
  NNodes = nNodes;
  NodalCoordinates = nodalCoordinates;
}



/* Node Search */

static AutoRes_Bucket2D NodeBucket;



void AutoRes_CreateEmptyNodeBucket2D 
(double nodalDistance,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY)
{
  int xDivisions, yDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY;
  int nAverageLinksPerCell = 2;

  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);

  xDivisions = (maxRangeX - minRangeX) / nodalDistance;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / nodalDistance;
  if (yDivisions < 1) yDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;

  AutoRes_Bucket2D_initialize 
    (&NodeBucket,
     minRangeX - offsetX, minRangeY - offsetY,
     maxRangeX + offsetX, maxRangeY + offsetY,
     xDivisions, yDivisions, 
     nAverageLinksPerCell);
}

void AutoRes_CreateNodeBucketAndAddNodes2D 
(double nodalDistance)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  int iNode;
  int xDivisions, yDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY;
  int nAverageLinksPerCell = 2;

  minRangeX = maxRangeX = NodalCoordinates[0 * N_DIMENSIONS + X];
  minRangeY = maxRangeY = NodalCoordinates[0 * N_DIMENSIONS + Y];
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (maxRangeX < NodalCoordinates[iNode * N_DIMENSIONS + X]) {
      maxRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (maxRangeY < NodalCoordinates[iNode * N_DIMENSIONS + Y]) {
      maxRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }

    if (NodalCoordinates[iNode * N_DIMENSIONS + X] < minRangeX) {
      minRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Y] < minRangeY) {
      minRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
  }
  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);

  xDivisions = (maxRangeX - minRangeX) / nodalDistance;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / nodalDistance;
  if (yDivisions < 1) yDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;

  AutoRes_Bucket2D_initialize 
    (&NodeBucket,
     minRangeX - offsetX, minRangeY - offsetY,
     maxRangeX + offsetX, maxRangeY + offsetY,
     xDivisions, yDivisions, 
     nAverageLinksPerCell);

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = NodalCoordinates[iNode * N_DIMENSIONS + X];
    double y = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    
    AutoRes_Bucket2D_addItemAt 
      (&NodeBucket,
       x, y,
       iNode);
  }
}

void AutoRes_AddLastNode2D (void)
{
  int lastNodeId = NNodes;
  double nodeX = NodalCoordinates[lastNodeId * N_DIMENSIONS + X];
  double nodeY = NodalCoordinates[lastNodeId * N_DIMENSIONS + Y];

  AutoRes_Bucket2D_addItemAt 
    (&NodeBucket,
     nodeX, nodeY, 
     lastNodeId);

  NNodes++;
}

AutoRes_Bucket2D *AutoRes_NodeBucket2D (void)
{
  return &NodeBucket;
}

int AutoRes_FindNodeAt2D 
(double x, double y)
{
  int cx, cy;
  double eps = AutoRes_GetTolerance ();

  AutoRes_Bucket2D_getIndexAt 
    (&NodeBucket, 
     x, y, 
     &cx, &cy);

  if (AutoRes_Bucket2D_isOutOfRange 
      (&NodeBucket, 
       cx, cy)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket2D_startLinkIndexIn 
	   (&NodeBucket, 
	    cx, cy);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket2D_nextLinkIndex 
	   (&NodeBucket, 
	    linkIndex)) {
      int nodeId = AutoRes_Bucket2D_itemId 
	(&NodeBucket, 
	 linkIndex);
      double nodeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      double nodeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      
      if (fabs (nodeX - x) < eps 
	  && fabs (nodeY - y) < eps) {
	return nodeId;
      }
    }
  }
  
  return -1;
}

void AutoRes_DeleteNodeBucket2D (void)
{
  AutoRes_Bucket2D_terminate 
    (&NodeBucket);
}



/* Element-Node Connectivity */

#define N_VERTEXS_PER_ELEMENT 3
#define N_EDGES_PER_ELEMENT 3
static int NElements;
static int NNodesPerElement;
static int *ElementNodeIds;   /* [NElements][NNodesPerElement] */



void AutoRes_SetTriangleMesh2D 
(int nElements,
 int nNodesPerElement,
 int elementNodeIds[])
{
  NElements = nElements;
  NNodesPerElement = nNodesPerElement;
  ElementNodeIds = elementNodeIds;
}

void AutoRes_GetEdgeNodesOfTriangleMesh2D 
(int elementId, int elementEdgeId,
 int *edgeNode0_OUT, int *edgeNode1_OUT)
{
  int elementNode0, elementNode1, elementNode2;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementEdgeId);
  assert(elementEdgeId < N_EDGES_PER_ELEMENT);

  elementNode0 = ElementNodeIds[elementId * NNodesPerElement + 0];
  elementNode1 = ElementNodeIds[elementId * NNodesPerElement + 1];
  elementNode2 = ElementNodeIds[elementId * NNodesPerElement + 2];

  switch (elementEdgeId) {
  case 0:
    *edgeNode0_OUT = elementNode1;
    *edgeNode1_OUT = elementNode2;
    break;
  case 1:
    *edgeNode0_OUT = elementNode2;
    *edgeNode1_OUT = elementNode0;
    break;
  case 2:
    *edgeNode0_OUT = elementNode0;
    *edgeNode1_OUT = elementNode1;
    break;
  default:
    assert(0);
    break;
  }
}



/* Edge Extraction */

static int HashSizeOfEdgeTable;
static AutoRes_ListTable EdgeTable;



static int HashFunctionOfEdge (int edgeId)
{
  int elementId = edgeId / N_EDGES_PER_ELEMENT;
  int elementEdgeId = edgeId % N_EDGES_PER_ELEMENT;
  int edgeNode0, edgeNode1;

  AutoRes_GetEdgeNodesOfTriangleMesh2D
    (elementId, elementEdgeId,
     &edgeNode0, &edgeNode1);

  return (edgeNode0 + edgeNode1) % HashSizeOfEdgeTable;
}

static int EdgesAreSame (int edgeIdA, int edgeIdB)
{
  int elementIdA = edgeIdA / N_EDGES_PER_ELEMENT;
  int elementEdgeIdA = edgeIdA % N_EDGES_PER_ELEMENT;
  int elementIdB = edgeIdB / N_EDGES_PER_ELEMENT;
  int elementEdgeIdB = edgeIdB % N_EDGES_PER_ELEMENT;
  int edgeNodeA0, edgeNodeA1;
  int edgeNodeB0, edgeNodeB1;

  AutoRes_GetEdgeNodesOfTriangleMesh2D
    (elementIdA, elementEdgeIdA,
     &edgeNodeA0, &edgeNodeA1);
  AutoRes_GetEdgeNodesOfTriangleMesh2D
    (elementIdB, elementEdgeIdB,
     &edgeNodeB0, &edgeNodeB1);

  if (edgeNodeA0 == edgeNodeB0
      && edgeNodeA1 == edgeNodeB1) return 1;
  if (edgeNodeA0 == edgeNodeB1
      && edgeNodeA1 == edgeNodeB0) return 1;
  return 0;
}

static void TryToAddEdge (int edgeIdToBeAdded)
{
  int hashValue = HashFunctionOfEdge (edgeIdToBeAdded);
  int theSameEdgeId;
  int linkIndex;

  theSameEdgeId = -1;
  for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	 (&EdgeTable, 
	  hashValue);
       linkIndex != -1;
       linkIndex = AutoRes_ListTable_nextLinkIndex 
	 (&EdgeTable, 
	  linkIndex)) {
    int edgeId = AutoRes_ListTable_itemId 
      (&EdgeTable, 
       linkIndex);
    
    if (EdgesAreSame (edgeId, edgeIdToBeAdded)) {
      theSameEdgeId = edgeId;
      break;
    }
  }

  if (theSameEdgeId == -1) {
    AutoRes_ListTable_addItemIn 
      (&EdgeTable,
       hashValue, edgeIdToBeAdded);
  } else {
    assert(HashFunctionOfEdge (theSameEdgeId) == hashValue);

    AutoRes_ListTable_removeItemIn 
      (&EdgeTable,
       hashValue, theSameEdgeId);
  }
}

void AutoRes_ExtractEdgesOfTriangleMesh2D 
(int maxEdges,
 int *nEdges_OUT,
 int elementIds_OUT[],
 int elementEdgeIds_OUT[])
{
  int nAverageLinksPerList = 1;
  int iElement;
  int iEdge;
  int count;
  int iList;
  int linkIndex;

  HashSizeOfEdgeTable = NElements;

  AutoRes_ListTable_initialize 
    (&EdgeTable,
     HashSizeOfEdgeTable, 
     nAverageLinksPerList);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iEdge = 0; iEdge < N_EDGES_PER_ELEMENT; iEdge++) {
      int edgeId = iElement * N_EDGES_PER_ELEMENT + iEdge;

      TryToAddEdge (edgeId);
    }
  }

  count = 0;
  for (iList = 0; 
       iList < AutoRes_ListTable_nLists 
	 (&EdgeTable); 
       iList++) {
    for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	   (&EdgeTable, 
	    iList);
	 linkIndex != -1;
	 linkIndex = AutoRes_ListTable_nextLinkIndex 
	   (&EdgeTable, 
	    linkIndex)) {
      count++;
    }
  }
  *nEdges_OUT = count;
  assert(*nEdges_OUT <= maxEdges);

  count = 0;
  for (iList = 0; 
       iList < AutoRes_ListTable_nLists 
	 (&EdgeTable); 
       iList++) {
    for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	   (&EdgeTable, 
	    iList);
	 linkIndex != -1;
	 linkIndex = AutoRes_ListTable_nextLinkIndex 
	   (&EdgeTable, 
	    linkIndex)) {
      int edgeId = AutoRes_ListTable_itemId 
	(&EdgeTable, 
	 linkIndex);
      int elementId = edgeId / N_EDGES_PER_ELEMENT;
      int elementEdgeId = edgeId % N_EDGES_PER_ELEMENT;

      elementIds_OUT[count] = elementId;
      elementEdgeIds_OUT[count] = elementEdgeId;
      count++;
    }
  }

  AutoRes_ListTable_terminate 
    (&EdgeTable);
}



/* Element Search */

static AutoRes_Bucket2D ElementBucket;



static double AverageElementSize (void)
{
  double result;
  int iElement;

  result = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId0 = ElementNodeIds[iElement * NNodesPerElement + 0];
    int nodeId1 = ElementNodeIds[iElement * NNodesPerElement + 1];
    int nodeId2 = ElementNodeIds[iElement * NNodesPerElement + 2];
    double x0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + X];
    double y0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Y];
    double x1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + X];
    double y1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Y];
    double x2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + X];
    double y2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Y];
    double edgeLength01 
      = sqrt ((x0 - x1) * (x0 - x1) + (y0 - y1) * (y0 - y1));
    double edgeLength12 
      = sqrt ((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
    double edgeLength20 
      = sqrt ((x2 - x0) * (x2 - x0) + (y2 - y0) * (y2 - y0));

    result += (edgeLength01 + edgeLength12 + edgeLength20) 
      / N_EDGES_PER_ELEMENT;
  }
  result /= NElements;

  return result;
}

static void CreateElementBucket (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  int iNode;
  double elementSize;
  int xDivisions, yDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY;
  int nAverageLinksPerCell = 10;

  minRangeX = maxRangeX = NodalCoordinates[0 * N_DIMENSIONS + X];
  minRangeY = maxRangeY = NodalCoordinates[0 * N_DIMENSIONS + Y];
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (maxRangeX < NodalCoordinates[iNode * N_DIMENSIONS + X]) {
      maxRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (maxRangeY < NodalCoordinates[iNode * N_DIMENSIONS + Y]) {
      maxRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }

    if (NodalCoordinates[iNode * N_DIMENSIONS + X] < minRangeX) {
      minRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Y] < minRangeY) {
      minRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
  }
  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);

  elementSize = AverageElementSize ();
  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;

  AutoRes_Bucket2D_initialize 
    (&ElementBucket,
     minRangeX - offsetX, minRangeY - offsetY,
     maxRangeX + offsetX, maxRangeY + offsetY,
     xDivisions, yDivisions, 
     nAverageLinksPerCell);
}

static void AddElementsIntoBucket (void)
{
  int iElement;
  int iNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId;
    double minRangeX, minRangeY;
    double maxRangeX, maxRangeY;
    
    nodeId = ElementNodeIds[iElement * NNodesPerElement + 0];
    minRangeX = maxRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
    minRangeY = maxRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
    for (iNode = 0; iNode < N_VERTEXS_PER_ELEMENT; iNode++) {
      nodeId = ElementNodeIds[iElement * NNodesPerElement + iNode];

      if (maxRangeX < NodalCoordinates[nodeId * N_DIMENSIONS + X]) {
	maxRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      }
      if (maxRangeY < NodalCoordinates[nodeId * N_DIMENSIONS + Y]) {
	maxRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      }

      if (NodalCoordinates[nodeId * N_DIMENSIONS + X] < minRangeX) {
	minRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      }
      if (NodalCoordinates[nodeId * N_DIMENSIONS + Y] < minRangeY) {
	minRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      }
    }
    
    AutoRes_Bucket2D_addItemWithin 
      (&ElementBucket,
       minRangeX, minRangeY,
       maxRangeX, maxRangeY,
       iElement);
  }
}

void AutoRes_CreateTriangleElementBucket2D (void)
{
  CreateElementBucket ();
  AddElementsIntoBucket ();
}

AutoRes_Bucket2D *AutoRes_TriangleElementBucket2D (void)
{
  return &ElementBucket;
}

static int ElementContainsPointAt 
(int iElement, 
 double x, double y,
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int nodeId0 = ElementNodeIds[iElement * NNodesPerElement + 0];
  int nodeId1 = ElementNodeIds[iElement * NNodesPerElement + 1];
  int nodeId2 = ElementNodeIds[iElement * NNodesPerElement + 2];
  double x0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + X];
  double y0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Y];
  double x1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + X];
  double y1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Y];
  double x2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + X];
  double y2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Y];
  double area = x1 * y2 + x0 * y1 + y0 * x2
    - y0 * x1 - y1 * x2 - y2 * x0;
  double eps = AutoRes_GetTolerance ();
  double eps2 = eps * eps;
  
  *L0_OUT = (x1 * y2 + x * y1 + y * x2
	     - y * x1 - y1 * x2 - y2 * x) / area;
  *L1_OUT = (x * y2 + x0 * y + y0 * x2
	     - y0 * x - y * x2 - y2 * x0) / area;
  *L2_OUT = (x1 * y + x0 * y1 + y0 * x
	     - y0 * x1 - y1 * x - y * x0) / area;
  
  if (*L0_OUT < -eps2) return 0;
  if (*L1_OUT < -eps2) return 0;
  if (*L2_OUT < -eps2) return 0;
  return 1;
}

int AutoRes_FindTriangleElementAt2D 
(double x, double y,
 double *L0_OUT, double *L1_OUT, double *L2_OUT)
{
  int cx, cy;

  AutoRes_Bucket2D_getIndexAt 
    (&ElementBucket, 
     x, y,
     &cx, &cy);
  
  if (AutoRes_Bucket2D_isOutOfRange 
      (&ElementBucket, 
       cx, cy)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket2D_startLinkIndexIn 
	   (&ElementBucket, 
	    cx, cy);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket2D_nextLinkIndex 
	   (&ElementBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket2D_itemId 
	(&ElementBucket, 
	 linkIndex);
      
      if (ElementContainsPointAt (elementId, x, y,
				  L0_OUT, L1_OUT, L2_OUT)) {
	return elementId;
      }
    }
  }
  
  return -1;
}

void AutoRes_DeleteTriangleElementBucket2D (void)
{
  AutoRes_Bucket2D_terminate 
    (&ElementBucket);
}


