/* autores_utility3d.c : 3-dimensional tetrahedron mesh */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_listtable.h"
#include "autores_bucket3d.h"
#include "autores_utility3d.h"



/* Nodal Coordinate */

#define N_DIMENSIONS 3
#define X 0
#define Y 1
#define Z 2
static int NNodes;
static double *NodalCoordinates;   /* [NNodes][N_DIMENSIONS] */



void AutoRes_SetNodalCoordinate3D 
(int nNodes,
 double nodalCoordinates[])
{
  NNodes = nNodes;
  NodalCoordinates = nodalCoordinates;
}



/* Node Search */

static AutoRes_Bucket3D NodeBucket;



void AutoRes_CreateEmptyNodeBucket3D 
(double nodalDistance,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ)
{
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
  int nAverageLinksPerCell = 2;

  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);
  assert(minRangeZ < maxRangeZ);

  xDivisions = (maxRangeX - minRangeX) / nodalDistance;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / nodalDistance;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / nodalDistance;
  if (zDivisions < 1) zDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&NodeBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
}

void AutoRes_CreateNodeBucketAndAddNodes3D 
(double nodalDistance)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  int iNode;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
  int nAverageLinksPerCell = 2;

  minRangeX = maxRangeX = NodalCoordinates[0 * N_DIMENSIONS + X];
  minRangeY = maxRangeY = NodalCoordinates[0 * N_DIMENSIONS + Y];
  minRangeZ = maxRangeZ = NodalCoordinates[0 * N_DIMENSIONS + Z];
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (maxRangeX < NodalCoordinates[iNode * N_DIMENSIONS + X]) {
      maxRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (maxRangeY < NodalCoordinates[iNode * N_DIMENSIONS + Y]) {
      maxRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
    if (maxRangeZ < NodalCoordinates[iNode * N_DIMENSIONS + Z]) {
      maxRangeZ = NodalCoordinates[iNode * N_DIMENSIONS + Z];
    }

    if (NodalCoordinates[iNode * N_DIMENSIONS + X] < minRangeX) {
      minRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Y] < minRangeY) {
      minRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Z] < minRangeZ) {
      minRangeZ = NodalCoordinates[iNode * N_DIMENSIONS + Z];
    }
  }
  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);
  assert(minRangeZ < maxRangeZ);

  xDivisions = (maxRangeX - minRangeX) / nodalDistance;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / nodalDistance;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / nodalDistance;
  if (zDivisions < 1) zDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&NodeBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);

  for (iNode = 0; iNode < NNodes; iNode++) {
    double x = NodalCoordinates[iNode * N_DIMENSIONS + X];
    double y = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    double z = NodalCoordinates[iNode * N_DIMENSIONS + Z];
    
    AutoRes_Bucket3D_addItemAt 
      (&NodeBucket,
       x, y, z,
       iNode);
  }
}

void AutoRes_AddLastNode3D (void)
{
  int lastNodeId = NNodes;
  double nodeX = NodalCoordinates[lastNodeId * N_DIMENSIONS + X];
  double nodeY = NodalCoordinates[lastNodeId * N_DIMENSIONS + Y];
  double nodeZ = NodalCoordinates[lastNodeId * N_DIMENSIONS + Z];

  AutoRes_Bucket3D_addItemAt 
    (&NodeBucket,
     nodeX, nodeY, nodeZ, 
     lastNodeId);

  NNodes++;
}

AutoRes_Bucket3D *AutoRes_NodeBucket3D (void)
{
  return &NodeBucket;
}

int AutoRes_FindNodeAt3D 
(double x, double y, double z)
{
  int cx, cy, cz;
  double eps = AutoRes_GetTolerance ();

  AutoRes_Bucket3D_getIndexAt 
    (&NodeBucket, 
     x, y, z, 
     &cx, &cy, &cz);

  if (AutoRes_Bucket3D_isOutOfRange 
      (&NodeBucket, 
       cx, cy, cz)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket3D_startLinkIndexIn 
	   (&NodeBucket, 
	    cx, cy, cz);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket3D_nextLinkIndex 
	   (&NodeBucket, 
	    linkIndex)) {
      int nodeId = AutoRes_Bucket3D_itemId 
	(&NodeBucket, 
	 linkIndex);
      double nodeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      double nodeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      double nodeZ = NodalCoordinates[nodeId * N_DIMENSIONS + Z];
      
      if (fabs (nodeX - x) < eps 
	  && fabs (nodeY - y) < eps
	  && fabs (nodeZ - z) < eps) {
	return nodeId;
      }
    }
  }
  
  return -1;
}

void AutoRes_DeleteNodeBucket3D (void)
{
  AutoRes_Bucket3D_terminate 
    (&NodeBucket);
}



/* Element-Node Connectivity */

#define N_VERTEXS_PER_ELEMENT 4
#define N_EDGES_PER_ELEMENT 6
#define N_FACES_PER_ELEMENT 4
static int NElements;
static int NNodesPerElement;
static int *ElementNodeIds;   /* [NElements][NNodesPerElement] */



void AutoRes_SetTetrahedronMesh3D 
(int nElements,
 int nNodesPerElement,
 int elementNodeIds[])
{
  NElements = nElements;
  NNodesPerElement = nNodesPerElement;
  ElementNodeIds = elementNodeIds;
}

void AutoRes_GetFaceNodesOfTetrahedronMesh3D 
(int elementId, int elementFaceId,
 int *faceNode0_OUT, int *faceNode1_OUT, int *faceNode2_OUT)
{
  int elementNode0, elementNode1, elementNode2, elementNode3;

  assert(0 <= elementId);
  assert(elementId < NElements);
  assert(0 <= elementFaceId);
  assert(elementFaceId < N_FACES_PER_ELEMENT);

  elementNode0 = ElementNodeIds[elementId * NNodesPerElement + 0];
  elementNode1 = ElementNodeIds[elementId * NNodesPerElement + 1];
  elementNode2 = ElementNodeIds[elementId * NNodesPerElement + 2];
  elementNode3 = ElementNodeIds[elementId * NNodesPerElement + 3];

  switch (elementFaceId) {
  case 0:
    *faceNode0_OUT = elementNode1;
    *faceNode1_OUT = elementNode2;
    *faceNode2_OUT = elementNode3;
    break;
  case 1:
    *faceNode0_OUT = elementNode2;
    *faceNode1_OUT = elementNode3;
    *faceNode2_OUT = elementNode0;
    break;
  case 2:
    *faceNode0_OUT = elementNode3;
    *faceNode1_OUT = elementNode0;
    *faceNode2_OUT = elementNode1;
    break;
  case 3:
    *faceNode0_OUT = elementNode0;
    *faceNode1_OUT = elementNode1;
    *faceNode2_OUT = elementNode2;
    break;
  default:
    assert(0);
    break;
  }
}



/* Face Extraction */

static int HashSizeOfFaceTable;
static AutoRes_ListTable FaceTable;



static int HashFunctionOfFace (int faceId)
{
  int elementId = faceId / N_FACES_PER_ELEMENT;
  int elementFaceId = faceId % N_FACES_PER_ELEMENT;
  int faceNode0, faceNode1, faceNode2;

  AutoRes_GetFaceNodesOfTetrahedronMesh3D
    (elementId, elementFaceId,
     &faceNode0, &faceNode1, &faceNode2);

  return (faceNode0 + faceNode1 + faceNode2) % HashSizeOfFaceTable;
}

static int FacesAreSame (int faceIdA, int faceIdB)
{
  int elementIdA = faceIdA / N_FACES_PER_ELEMENT;
  int elementFaceIdA = faceIdA % N_FACES_PER_ELEMENT;
  int elementIdB = faceIdB / N_FACES_PER_ELEMENT;
  int elementFaceIdB = faceIdB % N_FACES_PER_ELEMENT;
  int faceNodeA0, faceNodeA1, faceNodeA2;
  int faceNodeB0, faceNodeB1, faceNodeB2;

  AutoRes_GetFaceNodesOfTetrahedronMesh3D
    (elementIdA, elementFaceIdA,
     &faceNodeA0, &faceNodeA1, &faceNodeA2);
  AutoRes_GetFaceNodesOfTetrahedronMesh3D
    (elementIdB, elementFaceIdB,
     &faceNodeB0, &faceNodeB1, &faceNodeB2);

  if (faceNodeA0 == faceNodeB0
      && faceNodeA1 == faceNodeB1
      && faceNodeA2 == faceNodeB2) return 1;
  if (faceNodeA0 == faceNodeB1
      && faceNodeA1 == faceNodeB0
      && faceNodeA2 == faceNodeB2) return 1;
  if (faceNodeA0 == faceNodeB0
      && faceNodeA1 == faceNodeB2
      && faceNodeA2 == faceNodeB1) return 1;
  if (faceNodeA0 == faceNodeB2
      && faceNodeA1 == faceNodeB1
      && faceNodeA2 == faceNodeB0) return 1;
  if (faceNodeA0 == faceNodeB1
      && faceNodeA1 == faceNodeB2
      && faceNodeA2 == faceNodeB0) return 1;
  if (faceNodeA0 == faceNodeB2
      && faceNodeA1 == faceNodeB0
      && faceNodeA2 == faceNodeB1) return 1;
  return 0;
}

static void TryToAddFace (int faceIdToBeAdded)
{
  int hashValue = HashFunctionOfFace (faceIdToBeAdded);
  int theSameFaceId;
  int linkIndex;

  theSameFaceId = -1;
  for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	 (&FaceTable, 
	  hashValue);
       linkIndex != -1;
       linkIndex = AutoRes_ListTable_nextLinkIndex 
	 (&FaceTable, 
	  linkIndex)) {
    int faceId = AutoRes_ListTable_itemId 
      (&FaceTable, 
       linkIndex);
    
    if (FacesAreSame (faceId, faceIdToBeAdded)) {
      theSameFaceId = faceId;
      break;
    }
  }

  if (theSameFaceId == -1) {
    AutoRes_ListTable_addItemIn 
      (&FaceTable,
       hashValue, faceIdToBeAdded);
  } else {
    assert(HashFunctionOfFace (theSameFaceId) == hashValue);

    AutoRes_ListTable_removeItemIn 
      (&FaceTable,
       hashValue, theSameFaceId);
  }
}

void AutoRes_ExtractFacesOfTetrahedronMesh3D 
(int maxFaces,
 int *nFaces_OUT,
 int elementIds_OUT[],
 int elementFaceIds_OUT[])
{
  int nAverageLinksPerList = 1;
  int iElement;
  int iFace;
  int count;
  int iList;
  int linkIndex;

  HashSizeOfFaceTable = NElements;

  AutoRes_ListTable_initialize 
    (&FaceTable,
     HashSizeOfFaceTable, 
     nAverageLinksPerList);

  for (iElement = 0; iElement < NElements; iElement++) {
    for (iFace = 0; iFace < N_FACES_PER_ELEMENT; iFace++) {
      int faceId = iElement * N_FACES_PER_ELEMENT + iFace;

      TryToAddFace (faceId);
    }
  }

  count = 0;
  for (iList = 0; 
       iList < AutoRes_ListTable_nLists 
	 (&FaceTable); 
       iList++) {
    for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	   (&FaceTable, 
	    iList);
	 linkIndex != -1;
	 linkIndex = AutoRes_ListTable_nextLinkIndex 
	   (&FaceTable, 
	    linkIndex)) {
      count++;
    }
  }
  *nFaces_OUT = count;
  assert(*nFaces_OUT <= maxFaces);

  count = 0;
  for (iList = 0; 
       iList < AutoRes_ListTable_nLists 
	 (&FaceTable); 
       iList++) {
    for (linkIndex = AutoRes_ListTable_startLinkIndexIn 
	   (&FaceTable, 
	    iList);
	 linkIndex != -1;
	 linkIndex = AutoRes_ListTable_nextLinkIndex 
	   (&FaceTable, 
	    linkIndex)) {
      int faceId = AutoRes_ListTable_itemId 
	(&FaceTable, 
	 linkIndex);
      int elementId = faceId / N_FACES_PER_ELEMENT;
      int elementFaceId = faceId % N_FACES_PER_ELEMENT;

      elementIds_OUT[count] = elementId;
      elementFaceIds_OUT[count] = elementFaceId;
      count++;
    }
  }

  AutoRes_ListTable_terminate 
    (&FaceTable);
}



/* Element Search */

static AutoRes_Bucket3D ElementBucket;



static double AverageElementSize (void)
{
  double result;
  int iElement;

  result = 0.0;
  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId0 = ElementNodeIds[iElement * NNodesPerElement + 0];
    int nodeId1 = ElementNodeIds[iElement * NNodesPerElement + 1];
    int nodeId2 = ElementNodeIds[iElement * NNodesPerElement + 2];
    int nodeId3 = ElementNodeIds[iElement * NNodesPerElement + 3];
    double x0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + X];
    double y0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Y];
    double z0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Z];
    double x1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + X];
    double y1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Y];
    double z1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Z];
    double x2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + X];
    double y2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Y];
    double z2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Z];
    double x3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + X];
    double y3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + Y];
    double z3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + Z];
    double edgeLength01 = sqrt ((x0 - x1) * (x0 - x1) 
				+ (y0 - y1) * (y0 - y1) 
				+ (z0 - z1) * (z0 - z1));
    double edgeLength12 = sqrt ((x1 - x2) * (x1 - x2) 
				+ (y1 - y2) * (y1 - y2) 
				+ (z1 - z2) * (z1 - z2));
    double edgeLength20 = sqrt ((x2 - x0) * (x2 - x0) 
				+ (y2 - y0) * (y2 - y0)
				+ (z2 - z0) * (z2 - z0));
    double edgeLength03 = sqrt ((x0 - x3) * (x0 - x3) 
				+ (y0 - y3) * (y0 - y3) 
				+ (z0 - z3) * (z0 - z3));
    double edgeLength13 = sqrt ((x1 - x3) * (x1 - x3) 
				+ (y1 - y3) * (y1 - y3) 
				+ (z1 - z3) * (z1 - z3));
    double edgeLength23 = sqrt ((x2 - x3) * (x2 - x3) 
				+ (y2 - y3) * (y2 - y3) 
				+ (z2 - z3) * (z2 - z3));

    result += (edgeLength01 + edgeLength12 + edgeLength20
	       + edgeLength03 + edgeLength13 + edgeLength23) 
      / N_EDGES_PER_ELEMENT * 0.7;
  }
  result /= NElements;

  return result;
}

static void CreateElementBucket (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  int iNode;
  double elementSize;
  int xDivisions, yDivisions, zDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY, offsetZ;
#if 1
  int nAverageLinksPerCell = 50;
#else
  int nAverageLinksPerCell = 100;
#endif

  minRangeX = maxRangeX = NodalCoordinates[0 * N_DIMENSIONS + X];
  minRangeY = maxRangeY = NodalCoordinates[0 * N_DIMENSIONS + Y];
  minRangeZ = maxRangeZ = NodalCoordinates[0 * N_DIMENSIONS + Z];
  for (iNode = 0; iNode < NNodes; iNode++) {
    if (maxRangeX < NodalCoordinates[iNode * N_DIMENSIONS + X]) {
      maxRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (maxRangeY < NodalCoordinates[iNode * N_DIMENSIONS + Y]) {
      maxRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
    if (maxRangeZ < NodalCoordinates[iNode * N_DIMENSIONS + Z]) {
      maxRangeZ = NodalCoordinates[iNode * N_DIMENSIONS + Z];
    }

    if (NodalCoordinates[iNode * N_DIMENSIONS + X] < minRangeX) {
      minRangeX = NodalCoordinates[iNode * N_DIMENSIONS + X];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Y] < minRangeY) {
      minRangeY = NodalCoordinates[iNode * N_DIMENSIONS + Y];
    }
    if (NodalCoordinates[iNode * N_DIMENSIONS + Z] < minRangeZ) {
      minRangeZ = NodalCoordinates[iNode * N_DIMENSIONS + Z];
    }
  }
  assert(minRangeX < maxRangeX);
  assert(minRangeY < maxRangeY);
  assert(minRangeZ < maxRangeZ);

#if 0
  elementSize = AverageElementSize ();
#else
  elementSize = AverageElementSize () * 2;
#endif
  xDivisions = (maxRangeX - minRangeX) / elementSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / elementSize;
  if (yDivisions < 1) yDivisions = 1;
  zDivisions = (maxRangeZ - minRangeZ) / elementSize;
  if (zDivisions < 1) zDivisions = 1;
  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;
  offsetZ = (maxRangeZ - minRangeZ) / nOffsetDivisions;

  AutoRes_Bucket3D_initialize 
    (&ElementBucket,
     minRangeX - offsetX, minRangeY - offsetY, minRangeZ - offsetZ,
     maxRangeX + offsetX, maxRangeY + offsetY, maxRangeZ + offsetZ,
     xDivisions, yDivisions, zDivisions, 
     nAverageLinksPerCell);
}

static void AddElementsIntoBucket (void)
{
  int iElement;
  int iNode;

  for (iElement = 0; iElement < NElements; iElement++) {
    int nodeId;
    double minRangeX, minRangeY, minRangeZ;
    double maxRangeX, maxRangeY, maxRangeZ;
    
    nodeId = ElementNodeIds[iElement * NNodesPerElement + 0];
    minRangeX = maxRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
    minRangeY = maxRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
    minRangeZ = maxRangeZ = NodalCoordinates[nodeId * N_DIMENSIONS + Z];
    for (iNode = 0; iNode < N_VERTEXS_PER_ELEMENT; iNode++) {
      nodeId = ElementNodeIds[iElement * NNodesPerElement + iNode];

      if (maxRangeX < NodalCoordinates[nodeId * N_DIMENSIONS + X]) {
	maxRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      }
      if (maxRangeY < NodalCoordinates[nodeId * N_DIMENSIONS + Y]) {
	maxRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      }
      if (maxRangeZ < NodalCoordinates[nodeId * N_DIMENSIONS + Z]) {
	maxRangeZ = NodalCoordinates[nodeId * N_DIMENSIONS + Z];
      }

      if (NodalCoordinates[nodeId * N_DIMENSIONS + X] < minRangeX) {
	minRangeX = NodalCoordinates[nodeId * N_DIMENSIONS + X];
      }
      if (NodalCoordinates[nodeId * N_DIMENSIONS + Y] < minRangeY) {
	minRangeY = NodalCoordinates[nodeId * N_DIMENSIONS + Y];
      }
      if (NodalCoordinates[nodeId * N_DIMENSIONS + Z] < minRangeZ) {
	minRangeZ = NodalCoordinates[nodeId * N_DIMENSIONS + Z];
      }
    }
    
    AutoRes_Bucket3D_addItemWithin 
      (&ElementBucket,
       minRangeX, minRangeY, minRangeZ,
       maxRangeX, maxRangeY, maxRangeZ,
       iElement);
  }
}

void AutoRes_CreateTetrahedronElementBucket3D (void)
{
  CreateElementBucket ();
  AddElementsIntoBucket ();
}

AutoRes_Bucket3D *AutoRes_TetrahedronElementBucket3D (void)
{
  return &ElementBucket;
}

static double SignedVolume 
(double xk, double yk, double zk,
 double xl, double yl, double zl,
 double xm, double ym, double zm,
 double xn, double yn, double zn)
{
  double xlk = xl - xk;
  double ylk = yl - yk;
  double zlk = zl - zk;

  double xmk = xm - xk;
  double ymk = ym - yk;
  double zmk = zm - zk;
  
  double xnk = xn - xk;
  double ynk = yn - yk;
  double znk = zn - zk;
  
  double dmn = ymk * znk - zmk * ynk;
  double dln = ylk * znk - zlk * ynk;
  double dlm = ylk * zmk - zlk * ymk;
  
  return (xlk * dmn - xmk * dln + xnk * dlm) / 6;
}

static int ElementContainsPointAt 
(int iElement, 
 double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int nodeId0 = ElementNodeIds[iElement * NNodesPerElement + 0];
  int nodeId1 = ElementNodeIds[iElement * NNodesPerElement + 1];
  int nodeId2 = ElementNodeIds[iElement * NNodesPerElement + 2];
  int nodeId3 = ElementNodeIds[iElement * NNodesPerElement + 3];
  double x0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + X];
  double y0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Y];
  double z0 = NodalCoordinates[nodeId0 * N_DIMENSIONS + Z];
  double x1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + X];
  double y1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Y];
  double z1 = NodalCoordinates[nodeId1 * N_DIMENSIONS + Z];
  double x2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + X];
  double y2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Y];
  double z2 = NodalCoordinates[nodeId2 * N_DIMENSIONS + Z];
  double x3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + X];
  double y3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + Y];
  double z3 = NodalCoordinates[nodeId3 * N_DIMENSIONS + Z];
  double volume = SignedVolume (x0, y0, z0,
				x1, y1, z1,
				x2, y2, z2,
				x3, y3, z3);
  double eps = AutoRes_GetTolerance ();
  
  *L0_OUT = SignedVolume (x, y, z,
			  x1, y1, z1,
			  x2, y2, z2,
			  x3, y3, z3) / volume;
  *L1_OUT = SignedVolume (x0, y0, z0,
			  x, y, z,
			  x2, y2, z2,
			  x3, y3, z3) / volume;
  *L2_OUT = SignedVolume (x0, y0, z0,
			  x1, y1, z1,
			  x, y, z,
			  x3, y3, z3) / volume;
  *L3_OUT = SignedVolume (x0, y0, z0,
			  x1, y1, z1,
			  x2, y2, z2,
			  x, y, z) / volume;
  
  if (*L0_OUT < -eps) return 0;
  if (*L1_OUT < -eps) return 0;
  if (*L2_OUT < -eps) return 0;
  if (*L3_OUT < -eps) return 0;
  return 1;
}

int AutoRes_FindTetrahedronElementAt3D 
(double x, double y, double z,
 double *L0_OUT, double *L1_OUT, double *L2_OUT, double *L3_OUT)
{
  int cx, cy, cz;

  AutoRes_Bucket3D_getIndexAt 
    (&ElementBucket, 
     x, y, z,
     &cx, &cy, &cz);
  
  if (AutoRes_Bucket3D_isOutOfRange 
      (&ElementBucket, 
       cx, cy, cz)) {
    return -1;
  } else {
    int linkIndex;

    for (linkIndex = AutoRes_Bucket3D_startLinkIndexIn 
	   (&ElementBucket, 
	    cx, cy, cz);
	 linkIndex != -1;
	 linkIndex = AutoRes_Bucket3D_nextLinkIndex 
	   (&ElementBucket, 
	    linkIndex)) {
      int elementId = AutoRes_Bucket3D_itemId 
	(&ElementBucket, 
	 linkIndex);
      
      if (ElementContainsPointAt (elementId, x, y, z,
				  L0_OUT, L1_OUT, L2_OUT, L3_OUT)) {
	return elementId;
      }
    }
  }
  
  return -1;
}

void AutoRes_DeleteTetrahedronElementBucket3D (void)
{
  AutoRes_Bucket3D_terminate 
    (&ElementBucket);
}



