/* autores_bucket1d.c : 1-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_list_table.h"
#include "autores_bucket1d.h"



/* Bucket1D */

AUTORES_IMPLEMENT_HANDLE(AutoRes_Bucket1D);

#define LIST_ID_OFFSET 1
#define OUT_OF_RANGE_LIST_ID 0

AutoRes_Bucket1D *AutoRes_Bucket1D_New 
(double minRangeX, 
 double maxRangeX)
{
  AutoRes_Bucket1D *bucket;
  double tolerance = AutoRes_GetTolerance ();

  assert(minRangeX + tolerance < maxRangeX);

  bucket = AutoRes_Bucket1D_NewImplementation ();

  bucket->minRangeX = minRangeX;
  bucket->maxRangeX = maxRangeX;
  bucket->xCells = 0;

  bucket->listTable = AutoRes_ListTable_New ();
  assert(bucket->listTable != NULL);

  return bucket;
}

void AutoRes_Bucket1D_delete 
(AutoRes_Bucket1D *self)
{
  AutoRes_ListTable_delete (self->listTable);
  self->listTable = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_Bucket1D_reserve 
(AutoRes_Bucket1D *self,
 int xCells, 
 int nAverageLinksPerList)
{
  int flag;

  assert(0 < xCells);
  assert(0 < nAverageLinksPerList);
  
  self->xCells = xCells;

  flag = AutoRes_ListTable_reserve 
    (self->listTable, 
     xCells + LIST_ID_OFFSET, 
     nAverageLinksPerList);
  if (!flag) return 0;
  
  return 1;
}

void AutoRes_Bucket1D_getMinRange 
(AutoRes_Bucket1D *self,
 double *minRangeX_OUT)
{
  *minRangeX_OUT = self->minRangeX;
}

void AutoRes_Bucket1D_getMaxRange 
(AutoRes_Bucket1D *self,
 double *maxRangeX_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
}

void AutoRes_Bucket1D_getNCells 
(AutoRes_Bucket1D *self,
 int *xCells_OUT)
{
  *xCells_OUT = self->xCells;
}

void AutoRes_Bucket1D_getCellIdWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int *cx0_OUT, 
 int *cx1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX;
  double div0X;
  int cx0;
  double div1X;
  int cx1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  cx1 = (int)floor (div1X + tolerance);

  *cx0_OUT = cx0;
  *cx1_OUT = cx1;

  assert(*cx0_OUT <= *cx1_OUT);
}

void AutoRes_Bucket1D_getCellIdAt 
(AutoRes_Bucket1D *self,
 double x,
 int *cx_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX;
  double divX;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;

  divX = (x - self->minRangeX) / cellSizeX;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
}

int /* boolean */   
AutoRes_Bucket1D_isOutOfRange 
(AutoRes_Bucket1D *self,
 int cx)
{
  return cx < 0 || self->xCells <= cx;
}

int AutoRes_Bucket1D_startLinkIdIn 
(AutoRes_Bucket1D *self,
 int cx)
{
  assert(0 <= cx && cx < self->xCells);

  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     cx + LIST_ID_OFFSET);
}

int AutoRes_Bucket1D_startLinkIdOutOfRange 
(AutoRes_Bucket1D *self)
{
  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     OUT_OF_RANGE_LIST_ID);
}

int AutoRes_Bucket1D_nextLinkId 
(AutoRes_Bucket1D *self,
 int linkId)
{
  return AutoRes_ListTable_nextLinkId (self->listTable, linkId);
}

int AutoRes_Bucket1D_objectId 
(AutoRes_Bucket1D *self,
 int linkId)
{
  return AutoRes_ListTable_objectId (self->listTable, linkId);
}

int /* boolean */ 
AutoRes_Bucket1D_addObjectWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0;
  int cx1;
  int cx;
  int outOfRangeExists;
  int linkId;
  
  assert(minRangeX <= maxRangeX + tolerance);

  AutoRes_Bucket1D_getCellIdWithin 
    (self,
     minRangeX, 
     maxRangeX, 
     &cx0, 
     &cx1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    if (AutoRes_Bucket1D_isOutOfRange 
	(self,
	 cx)) {
      outOfRangeExists = 1;
    } else {
      linkId = AutoRes_ListTable_addObject
	(self->listTable, 
	 cx + LIST_ID_OFFSET, 
	 objectId);
      if (linkId == -1) return 0;
    }
  }
  if (outOfRangeExists) {
    linkId = AutoRes_ListTable_addObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
    if (linkId == -1) return 0;
  }

  return 1;
}

int AutoRes_Bucket1D_addObjectAt 
(AutoRes_Bucket1D *self,
 double x,
 int objectId)
{
  int flag;

  flag = AutoRes_Bucket1D_addObjectWithin 
    (self,
     x, 
     x, 
     objectId);
  if (!flag) return 0;

  return 1;
}

void AutoRes_Bucket1D_removeObjectWithin 
(AutoRes_Bucket1D *self,
 double minRangeX,
 double maxRangeX,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0;
  int cx1;
  int cx;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);

  AutoRes_Bucket1D_getCellIdWithin 
    (self,
     minRangeX, 
     maxRangeX, 
     &cx0, 
     &cx1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    if (AutoRes_Bucket1D_isOutOfRange 
	(self,
	 cx)) {
      outOfRangeExists = 1;
    } else {
      AutoRes_ListTable_removeObject
	(self->listTable, 
	 cx + LIST_ID_OFFSET, 
	 objectId);
    }
  }
  if (outOfRangeExists) {
    AutoRes_ListTable_removeObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
  }
}

void AutoRes_Bucket1D_removeObjectAt 
(AutoRes_Bucket1D *self,
 double x,
 int objectId)
{
  AutoRes_Bucket1D_removeObjectWithin 
    (self,
     x, 
     x, 
     objectId);
}

