/* autores_bucket2d.c : 2-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_list_table.h"
#include "autores_bucket2d.h"



/* Bucket2D */

AUTORES_IMPLEMENT_HANDLE(AutoRes_Bucket2D);

#define LIST_ID_OFFSET 1
#define OUT_OF_RANGE_LIST_ID 0

AutoRes_Bucket2D *AutoRes_Bucket2D_New 
(double minRangeX, double minRangeY, 
 double maxRangeX, double maxRangeY)
{
  AutoRes_Bucket2D *bucket;
  double tolerance = AutoRes_GetTolerance ();

  assert(minRangeX + tolerance < maxRangeX);
  assert(minRangeY + tolerance < maxRangeY);

  bucket = AutoRes_Bucket2D_NewImplementation ();

  bucket->minRangeX = minRangeX;
  bucket->minRangeY = minRangeY;
  bucket->maxRangeX = maxRangeX;
  bucket->maxRangeY = maxRangeY;
  bucket->xCells = 0;
  bucket->yCells = 0;

  bucket->listTable = AutoRes_ListTable_New ();
  assert(bucket->listTable != NULL);

  return bucket;
}

void AutoRes_Bucket2D_delete 
(AutoRes_Bucket2D *self)
{
  AutoRes_ListTable_delete (self->listTable);
  self->listTable = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_Bucket2D_reserve 
(AutoRes_Bucket2D *self,
 int xCells, int yCells, 
 int nAverageLinksPerList)
{
  int flag;

  assert(0 < xCells);
  assert(0 < yCells);
  assert(0 < nAverageLinksPerList);
  
  self->xCells = xCells;
  self->yCells = yCells;

  flag = AutoRes_ListTable_reserve 
    (self->listTable, 
     xCells * yCells + LIST_ID_OFFSET, 
     nAverageLinksPerList);
  if (!flag) return 0;
  
  return 1;
}

void AutoRes_Bucket2D_getMinRange 
(AutoRes_Bucket2D *self,
 double *minRangeX_OUT, double *minRangeY_OUT)
{
  *minRangeX_OUT = self->minRangeX;
  *minRangeY_OUT = self->minRangeY;
}

void AutoRes_Bucket2D_getMaxRange 
(AutoRes_Bucket2D *self,
 double *maxRangeX_OUT, double *maxRangeY_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
  *maxRangeY_OUT = self->maxRangeY;
}

void AutoRes_Bucket2D_getNCells 
(AutoRes_Bucket2D *self,
 int *xCells_OUT, int *yCells_OUT)
{
  *xCells_OUT = self->xCells;
  *yCells_OUT = self->yCells;
}

void AutoRes_Bucket2D_getCellIdWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int *cx0_OUT, int *cy0_OUT, 
 int *cx1_OUT, int *cy1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY;
  double div0X, div0Y;
  int cx0, cy0;
  double div1X, div1Y;
  int cx1, cy1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  div0Y = (minRangeY - self->minRangeY) / cellSizeY;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;
  cy0 = (int)floor (div0Y + tolerance);
  if (fabs (div0Y - cy0) < tolerance) cy0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  div1Y = (maxRangeY - self->minRangeY) / cellSizeY;
  cx1 = (int)floor (div1X + tolerance);
  cy1 = (int)floor (div1Y + tolerance);

  *cx0_OUT = cx0;
  *cy0_OUT = cy0;
  *cx1_OUT = cx1;
  *cy1_OUT = cy1;

  assert(*cx0_OUT <= *cx1_OUT);
  assert(*cy0_OUT <= *cy1_OUT);
}

void AutoRes_Bucket2D_getCellIdAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int *cx_OUT, int *cy_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY;
  double divX, divY;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;

  divX = (x - self->minRangeX) / cellSizeX;
  divY = (y - self->minRangeY) / cellSizeY;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
  *cy_OUT = (int)floor (divY + tolerance);
  if (fabs (y - self->maxRangeY) < tolerance) *cy_OUT = self->yCells - 1;
}

int /* boolean */   
AutoRes_Bucket2D_isOutOfRange 
(AutoRes_Bucket2D *self,
 int cx, int cy)
{
  return cx < 0 || self->xCells <= cx 
    || cy < 0 || self->yCells <= cy;
}

int AutoRes_Bucket2D_startLinkIdIn 
(AutoRes_Bucket2D *self,
 int cx, int cy)
{
  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells);

  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     cx * self->yCells + cy + LIST_ID_OFFSET);
}

int AutoRes_Bucket2D_startLinkIdOutOfRange 
(AutoRes_Bucket2D *self)
{
  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     OUT_OF_RANGE_LIST_ID);
}

int AutoRes_Bucket2D_nextLinkId 
(AutoRes_Bucket2D *self,
 int linkId)
{
  return AutoRes_ListTable_nextLinkId (self->listTable, linkId);
}

int AutoRes_Bucket2D_objectId 
(AutoRes_Bucket2D *self,
 int linkId)
{
  return AutoRes_ListTable_objectId (self->listTable, linkId);
}

int /* boolean */   
AutoRes_Bucket2D_addObjectWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0;
  int cx1, cy1;
  int cx, cy;
  int outOfRangeExists;
  int linkId;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);

  AutoRes_Bucket2D_getCellIdWithin 
    (self,
     minRangeX, minRangeY, 
     maxRangeX, maxRangeY, 
     &cx0, &cy0, 
     &cx1, &cy1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      if (AutoRes_Bucket2D_isOutOfRange 
	  (self,
	   cx, cy)) {
	outOfRangeExists = 1;
      } else {
	linkId = AutoRes_ListTable_addObject
	  (self->listTable, 
	   cx * self->yCells + cy + LIST_ID_OFFSET, 
	   objectId);
	if (linkId == -1) return 0;
      }
    }
  }
  if (outOfRangeExists) {
    linkId = AutoRes_ListTable_addObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
    if (linkId == -1) return 0;
  }

  return 1;
}

int /* boolean */   
AutoRes_Bucket2D_addObjectAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int objectId)
{
  int flag;

  flag = AutoRes_Bucket2D_addObjectWithin 
    (self,
     x, y, 
     x, y, 
     objectId);
  if (!flag) return 0;

  return 1;
}

void AutoRes_Bucket2D_removeObjectWithin 
(AutoRes_Bucket2D *self,
 double minRangeX, double minRangeY,
 double maxRangeX, double maxRangeY,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0;
  int cx1, cy1;
  int cx, cy;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);

  AutoRes_Bucket2D_getCellIdWithin 
    (self,
     minRangeX, minRangeY, 
     maxRangeX, maxRangeY, 
     &cx0, &cy0, 
     &cx1, &cy1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      if (AutoRes_Bucket2D_isOutOfRange 
	  (self,
	   cx, cy)) {
	outOfRangeExists = 1;
      } else {
	AutoRes_ListTable_removeObject
	  (self->listTable, 
	   cx * self->yCells + cy + LIST_ID_OFFSET, 
	   objectId);
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_ListTable_removeObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
  }
}

void AutoRes_Bucket2D_removeObjectAt 
(AutoRes_Bucket2D *self,
 double x, double y,
 int objectId)
{
  AutoRes_Bucket2D_removeObjectWithin 
    (self,
     x, y, 
     x, y, 
     objectId);
}

