/* autores_bucket3d.c : 3-dimensional bucket */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_list_table.h"
#include "autores_bucket3d.h"



/* Bucket3D */

AUTORES_IMPLEMENT_HANDLE(AutoRes_Bucket3D);

#define LIST_ID_OFFSET 1
#define OUT_OF_RANGE_LIST_ID 0

AutoRes_Bucket3D *AutoRes_Bucket3D_New 
(double minRangeX, double minRangeY, double minRangeZ, 
 double maxRangeX, double maxRangeY, double maxRangeZ)
{
  AutoRes_Bucket3D *bucket;
  double tolerance = AutoRes_GetTolerance ();

  assert(minRangeX + tolerance < maxRangeX);
  assert(minRangeY + tolerance < maxRangeY);
  assert(minRangeZ + tolerance < maxRangeZ);

  bucket = AutoRes_Bucket3D_NewImplementation ();

  bucket->minRangeX = minRangeX;
  bucket->minRangeY = minRangeY;
  bucket->minRangeZ = minRangeZ;
  bucket->maxRangeX = maxRangeX;
  bucket->maxRangeY = maxRangeY;
  bucket->maxRangeZ = maxRangeZ;
  bucket->xCells = 0;
  bucket->yCells = 0;
  bucket->zCells = 0;

  bucket->listTable = AutoRes_ListTable_New ();
  assert(bucket->listTable != NULL);

  return bucket;
}

void AutoRes_Bucket3D_delete 
(AutoRes_Bucket3D *self)
{
  AutoRes_ListTable_delete (self->listTable);
  self->listTable = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_Bucket3D_reserve 
(AutoRes_Bucket3D *self,
 int xCells, int yCells, int zCells, 
 int nAverageLinksPerList)
{
  int flag;

  assert(0 < xCells);
  assert(0 < yCells);
  assert(0 < zCells);
  assert(0 < nAverageLinksPerList);
  
  self->xCells = xCells;
  self->yCells = yCells;
  self->zCells = zCells;

  flag = AutoRes_ListTable_reserve 
    (self->listTable, 
     xCells * yCells * zCells + LIST_ID_OFFSET, 
     nAverageLinksPerList);
  if (!flag) return 0;
  
  return 1;
}

void AutoRes_Bucket3D_getMinRange 
(AutoRes_Bucket3D *self,
 double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT)
{
  *minRangeX_OUT = self->minRangeX;
  *minRangeY_OUT = self->minRangeY;
  *minRangeZ_OUT = self->minRangeZ;
}

void AutoRes_Bucket3D_getMaxRange 
(AutoRes_Bucket3D *self,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  *maxRangeX_OUT = self->maxRangeX;
  *maxRangeY_OUT = self->maxRangeY;
  *maxRangeZ_OUT = self->maxRangeZ;
}

void AutoRes_Bucket3D_getNCells 
(AutoRes_Bucket3D *self,
 int *xCells_OUT, int *yCells_OUT, int *zCells_OUT)
{
  *xCells_OUT = self->xCells;
  *yCells_OUT = self->yCells;
  *zCells_OUT = self->zCells;
}

void AutoRes_Bucket3D_getCellIdWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int *cx0_OUT, int *cy0_OUT, int *cz0_OUT, 
 int *cx1_OUT, int *cy1_OUT, int *cz1_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY, cellSizeZ;
  double div0X, div0Y, div0Z;
  int cx0, cy0, cz0;
  double div1X, div1Y, div1Z;
  int cx1, cy1, cz1;

  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;
  cellSizeZ = (self->maxRangeZ - self->minRangeZ) / self->zCells;

  div0X = (minRangeX - self->minRangeX) / cellSizeX;
  div0Y = (minRangeY - self->minRangeY) / cellSizeY;
  div0Z = (minRangeZ - self->minRangeZ) / cellSizeZ;
  cx0 = (int)floor (div0X + tolerance);
  if (fabs (div0X - cx0) < tolerance) cx0--;
  cy0 = (int)floor (div0Y + tolerance);
  if (fabs (div0Y - cy0) < tolerance) cy0--;
  cz0 = (int)floor (div0Z + tolerance);
  if (fabs (div0Z - cz0) < tolerance) cz0--;

  div1X = (maxRangeX - self->minRangeX) / cellSizeX;
  div1Y = (maxRangeY - self->minRangeY) / cellSizeY;
  div1Z = (maxRangeZ - self->minRangeZ) / cellSizeZ;
  cx1 = (int)floor (div1X + tolerance);
  cy1 = (int)floor (div1Y + tolerance);
  cz1 = (int)floor (div1Z + tolerance);

  *cx0_OUT = cx0;
  *cy0_OUT = cy0;
  *cz0_OUT = cz0;
  *cx1_OUT = cx1;
  *cy1_OUT = cy1;
  *cz1_OUT = cz1;

  assert(*cx0_OUT <= *cx1_OUT);
  assert(*cy0_OUT <= *cy1_OUT);
  assert(*cz0_OUT <= *cz1_OUT);
}

void AutoRes_Bucket3D_getCellIdAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int *cx_OUT, int *cy_OUT, int *cz_OUT)
{
  double tolerance = AutoRes_GetTolerance ();
  double cellSizeX, cellSizeY, cellSizeZ;
  double divX, divY, divZ;
  
  cellSizeX = (self->maxRangeX - self->minRangeX) / self->xCells;
  cellSizeY = (self->maxRangeY - self->minRangeY) / self->yCells;
  cellSizeZ = (self->maxRangeZ - self->minRangeZ) / self->zCells;

  divX = (x - self->minRangeX) / cellSizeX;
  divY = (y - self->minRangeY) / cellSizeY;
  divZ = (z - self->minRangeZ) / cellSizeZ;
  *cx_OUT = (int)floor (divX + tolerance);
  if (fabs (x - self->maxRangeX) < tolerance) *cx_OUT = self->xCells - 1;
  *cy_OUT = (int)floor (divY + tolerance);
  if (fabs (y - self->maxRangeY) < tolerance) *cy_OUT = self->yCells - 1;
  *cz_OUT = (int)floor (divZ + tolerance);
  if (fabs (z - self->maxRangeZ) < tolerance) *cz_OUT = self->zCells - 1;
}

int /* boolean */    
AutoRes_Bucket3D_isOutOfRange 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz)
{
  return cx < 0 || self->xCells <= cx 
    || cy < 0 || self->yCells <= cy
    || cz < 0 || self->zCells <= cz;
}

int AutoRes_Bucket3D_startLinkIdIn 
(AutoRes_Bucket3D *self,
 int cx, int cy, int cz)
{
  assert(0 <= cx && cx < self->xCells
	 && 0 <= cy && cy < self->yCells
	 && 0 <= cz && cz < self->zCells);

  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     cx * self->yCells * self->zCells 
     + cy * self->zCells + cz + LIST_ID_OFFSET);
}

int AutoRes_Bucket3D_startLinkIdOutOfRange 
(AutoRes_Bucket3D *self)
{
  return AutoRes_ListTable_startLinkId
    (self->listTable, 
     OUT_OF_RANGE_LIST_ID);
}

int AutoRes_Bucket3D_nextLinkId 
(AutoRes_Bucket3D *self,
 int linkId)
{
  return AutoRes_ListTable_nextLinkId (self->listTable, linkId);
}

int AutoRes_Bucket3D_objectId 
(AutoRes_Bucket3D *self,
 int linkId)
{
  return AutoRes_ListTable_objectId (self->listTable, linkId);
}

int /* boolean */    
AutoRes_Bucket3D_addObjectWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0, cz0;
  int cx1, cy1, cz1;
  int cx, cy, cz;
  int outOfRangeExists;
  int linkId;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);
  assert(minRangeZ <= maxRangeZ + tolerance);

  AutoRes_Bucket3D_getCellIdWithin 
    (self,
     minRangeX, minRangeY, minRangeZ, 
     maxRangeX, maxRangeY, maxRangeZ, 
     &cx0, &cy0, &cz0, 
     &cx1, &cy1, &cz1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      for (cz = cz0; cz <= cz1; cz++) {
	if (AutoRes_Bucket3D_isOutOfRange 
	    (self,
	     cx, cy, cz)) {
	  outOfRangeExists = 1;
	} else {
	  linkId = AutoRes_ListTable_addObject
	    (self->listTable, 
	     cx * self->yCells * self->zCells 
	     + cy * self->zCells + cz + LIST_ID_OFFSET, 
	     objectId);
	  if (linkId == -1) return 0;
	}
      }
    }
  }
  if (outOfRangeExists) {
    linkId = AutoRes_ListTable_addObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
    if (linkId == -1) return 0;
  }

  return 1;
}

int /* boolean */    
AutoRes_Bucket3D_addObjectAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int objectId)
{
  int flag;

  flag = AutoRes_Bucket3D_addObjectWithin 
    (self,
     x, y, z, 
     x, y, z, 
     objectId);
  if (!flag) return 0;

  return 1;
}

void AutoRes_Bucket3D_removeObjectWithin 
(AutoRes_Bucket3D *self,
 double minRangeX, double minRangeY, double minRangeZ,
 double maxRangeX, double maxRangeY, double maxRangeZ,
 int objectId)
{
  double tolerance = AutoRes_GetTolerance ();
  int cx0, cy0, cz0;
  int cx1, cy1, cz1;
  int cx, cy, cz;
  int outOfRangeExists;
  
  assert(minRangeX <= maxRangeX + tolerance);
  assert(minRangeY <= maxRangeY + tolerance);
  assert(minRangeZ <= maxRangeZ + tolerance);

  AutoRes_Bucket3D_getCellIdWithin 
    (self,
     minRangeX, minRangeY, minRangeZ, 
     maxRangeX, maxRangeY, maxRangeZ, 
     &cx0, &cy0, &cz0, 
     &cx1, &cy1, &cz1);

  outOfRangeExists = 0;
  for (cx = cx0; cx <= cx1; cx++) {
    for (cy = cy0; cy <= cy1; cy++) {
      for (cz = cz0; cz <= cz1; cz++) {
	if (AutoRes_Bucket3D_isOutOfRange 
	    (self,
	     cx, cy, cz)) {
	  outOfRangeExists = 1;
	} else {
	  AutoRes_ListTable_removeObject
	    (self->listTable, 
	     cx * self->yCells * self->zCells 
	     + cy * self->zCells + cz + LIST_ID_OFFSET, 
	     objectId);
	}
      }
    }
  }
  if (outOfRangeExists) {
    AutoRes_ListTable_removeObject
      (self->listTable, 
       OUT_OF_RANGE_LIST_ID, 
       objectId);
  }
}

void AutoRes_Bucket3D_removeObjectAt 
(AutoRes_Bucket3D *self,
 double x, double y, double z,
 int objectId)
{
  AutoRes_Bucket3D_removeObjectWithin 
    (self,
     x, y, z, 
     x, y, z, 
     objectId);
}

