/* autores_free_list.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_free_list.h"



/* Free List Collection */

#define COLUMN_NAME_NEXT_RECORD_ID "AutoRes:nextRecordId"

#define NULL_RECORD_ID -1
#define USE_RECORD_ID -2

AUTORES_IMPLEMENT_HANDLE(AutoRes_FreeListCollection);

AutoRes_FreeListCollection *AutoRes_FreeListCollection_New 
(AutoRes_Table *table)
{
  AutoRes_FreeListCollection *collection;

  collection = AutoRes_FreeListCollection_NewImplementation ();

  collection->table = table;
  collection->nextRecordIdColumn = NULL;
  collection->startRecordId = NULL_RECORD_ID;

  return collection;
}

void AutoRes_FreeListCollection_delete 
(AutoRes_FreeListCollection *self)
{
  if (self->nextRecordIdColumn != NULL) {
    AutoRes_Table_removeColumn 
      (self->table, 
       AutoRes_Table_columnIdOfColumn 
       (self->table, self->nextRecordIdColumn));
    self->nextRecordIdColumn = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_FreeListCollection_makeColumn 
(AutoRes_FreeListCollection *self)
{
  int flag;

  assert(self->nextRecordIdColumn == NULL);

  self->nextRecordIdColumn = AutoRes_IntegerColumn_New 
    (self->table, COLUMN_NAME_NEXT_RECORD_ID);
  assert(self->nextRecordIdColumn != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->table, self->nextRecordIdColumn);
  if (!flag) return 0;

  assert(self->nextRecordIdColumn != NULL);

  return 1;
}

void AutoRes_FreeListCollection_initializeColumn 
(AutoRes_FreeListCollection *self)
{
  int nRecords = AutoRes_Table_nRecords (self->table);
  int iRecord;

  assert(self->nextRecordIdColumn != NULL);

  self->startRecordId = NULL_RECORD_ID;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    AUTORES_I_SET(self->nextRecordIdColumn, iRecord, USE_RECORD_ID);
  }
}

int AutoRes_FreeListCollection_addRecord 
(AutoRes_FreeListCollection *self)
{
  int newRecordId;

  assert(0 < AutoRes_Table_nReservedRecords (self->table));
  assert(self->nextRecordIdColumn != NULL);

  if (self->startRecordId != NULL_RECORD_ID) {
    newRecordId = self->startRecordId;
    self->startRecordId = AUTORES_I(self->nextRecordIdColumn, newRecordId);
  } else {
    newRecordId = AutoRes_Table_addRecord (self->table);
    if (newRecordId == -1) return -1;
  }
  assert(0 <= newRecordId);
  AUTORES_I_SET(self->nextRecordIdColumn, newRecordId, USE_RECORD_ID);

  return newRecordId;
}

void AutoRes_FreeListCollection_removeRecord 
(AutoRes_FreeListCollection *self,
 int recordId)
{
  assert(self->nextRecordIdColumn != NULL);
  assert(AUTORES_I(self->nextRecordIdColumn, recordId) == USE_RECORD_ID);

  AUTORES_I_SET(self->nextRecordIdColumn, recordId, self->startRecordId);
  self->startRecordId = recordId;
}

static void PackIntegerColumn 
(AutoRes_Table *table,
 AutoRes_IntegerColumn *column,
 AutoRes_IntegerColumn *nextRecordIdColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int iRecord;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int nextRecordId = AUTORES_I(nextRecordIdColumn, iRecord);
    
    if (nextRecordId == USE_RECORD_ID) {
      int value = AUTORES_I(column, iRecord);

      AUTORES_I_SET(column, count, value);
      count++;
    }
  }
}

static void PackInteger1DColumn 
(AutoRes_Table *table,
 AutoRes_Integer1DColumn *column,
 AutoRes_IntegerColumn *nextRecordIdColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int size = AutoRes_Integer1DColumn_size (column);
  int iRecord;
  int index;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int nextRecordId = AUTORES_I(nextRecordIdColumn, iRecord);
    
    if (nextRecordId == USE_RECORD_ID) {
      for (index = 0; index < size; index++) {
	int value = AUTORES_I1D(column, iRecord, index);

	AUTORES_I1D_SET(column, count, index, value);
      }
      count++;
    }
  }
}

static void PackRealColumn 
(AutoRes_Table *table,
 AutoRes_RealColumn *column,
 AutoRes_IntegerColumn *nextRecordIdColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int iRecord;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int nextRecordId = AUTORES_I(nextRecordIdColumn, iRecord);
    
    if (nextRecordId == USE_RECORD_ID) {
      double value = AUTORES_R(column, iRecord);

      AUTORES_R_SET(column, count, value);
      count++;
    }
  }
}

static void PackReal1DColumn 
(AutoRes_Table *table,
 AutoRes_Real1DColumn *column,
 AutoRes_IntegerColumn *nextRecordIdColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int size = AutoRes_Real1DColumn_size (column);
  int iRecord;
  int index;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int nextRecordId = AUTORES_I(nextRecordIdColumn, iRecord);
    
    if (nextRecordId == USE_RECORD_ID) {
      for (index = 0; index < size; index++) {
	double value = AUTORES_R1D(column, iRecord, index);

	AUTORES_R1D_SET(column, count, index, value);
      }
      count++;
    }
  }
}

void AutoRes_FreeListCollection_pack 
(AutoRes_FreeListCollection *self)
{
  int nRecords;
  int nColumns;
  int iColumn;
  int iRecord;
  int count;

  assert(self->nextRecordIdColumn != NULL);

  nRecords = AutoRes_Table_nRecords (self->table);
  nColumns = AutoRes_Table_nColumns (self->table);

  for (iColumn = 0; iColumn < nColumns; iColumn++) {
    void *column = AutoRes_Table_column (self->table, iColumn);
    AutoRes_ColumnType columnType 
      = AutoRes_Table_columnType (self->table, iColumn);

    if (column == (void *)self->nextRecordIdColumn) {
      continue;
    }

    switch (columnType) {
    case AUTORES_COLUMN_INTEGER:
      PackIntegerColumn 
	(self->table, (AutoRes_IntegerColumn *)column, 
	 self->nextRecordIdColumn);
      break;
    case AUTORES_COLUMN_INTEGER1D:
      PackInteger1DColumn 
	(self->table, (AutoRes_Integer1DColumn *)column, 
	 self->nextRecordIdColumn);
      break;
    case AUTORES_COLUMN_REAL:
      PackRealColumn 
	(self->table, (AutoRes_RealColumn *)column, 
	 self->nextRecordIdColumn);
      break;
    case AUTORES_COLUMN_REAL1D:
      PackReal1DColumn 
	(self->table, (AutoRes_Real1DColumn *)column, 
	 self->nextRecordIdColumn);
      break;
    default:
      assert(0);
      break;
    }
  }

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int nextRecordId = AUTORES_I(self->nextRecordIdColumn, iRecord);

    if (nextRecordId == USE_RECORD_ID) {
      count++;
    }
  }
  for (iRecord = 0; iRecord < count; iRecord++) {
    AUTORES_I_SET(self->nextRecordIdColumn, iRecord, USE_RECORD_ID);
  }
  for (iRecord = count; iRecord < nRecords; iRecord++) {
    AUTORES_I_SET(self->nextRecordIdColumn, iRecord, NULL_RECORD_ID);
  }
  self->startRecordId = NULL_RECORD_ID;
  AutoRes_Table_packNRecords (self->table, count);
}

AutoRes_Table *AutoRes_FreeListCollection_table 
(AutoRes_FreeListCollection *self)
{
  return self->table;
}

int /* boolean */
AutoRes_FreeListCollection_recordIsUsed 
(AutoRes_FreeListCollection *self,
 int recordId)
{
  return AUTORES_I(self->nextRecordIdColumn, recordId) == USE_RECORD_ID;
}



