/* autores_integer_column.c : Table-based Memory Manager : Integer Column */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"



/* Table-based Memory Manager : Integer Column */



/* IntegerColumn */

AUTORES_IMPLEMENT_HANDLE(AutoRes_IntegerColumn);

AutoRes_IntegerColumn *AutoRes_IntegerColumn_New 
(AutoRes_Table *table,
 const char *name)
{
  AutoRes_IntegerColumn *column;

  assert(0 < strlen (name));
  assert(strlen (name) < AUTORES_MAX_COLUMN_NAME_LENGTH);

  column = AutoRes_IntegerColumn_NewImplementation ();

  column->table = table;
  column->block = NULL;

  column->name = (char *)malloc (strlen (name) + 1);
  assert(column->name != NULL);
  strcpy (column->name, name);

  return column;
}

void AutoRes_IntegerColumn_delete 
(AutoRes_IntegerColumn *self)
{
  free (self->name);
  self->name = NULL;

  if (self->block != NULL) {
    AutoRes_Memory_FreeBlock (self->block);
    self->block = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_IntegerColumn_allocate 
(AutoRes_IntegerColumn *self)
{
  int nReservedRecords;
  int memorySize;
  int flag;

  nReservedRecords = AutoRes_Table_nReservedRecords (self->table);
  memorySize = sizeof (int) * nReservedRecords;
  
  if (self->block != NULL) {
    flag = AutoRes_Memory_ReallocateBlock (self->block, memorySize);
    if (!flag) return 0;
  } else {
    self->block = AutoRes_Memory_AllocateBlock (memorySize);
    if (self->block == NULL) return 0;
  }
  return 1;
}

void AutoRes_IntegerColumn_setLockFlag 
(AutoRes_IntegerColumn *self,
 int flag)
{
  assert(self->block != NULL);

  AutoRes_Memory_SetLockFlagOnBlock (self->block, flag);
}

void AutoRes_IntegerColumn_setValue 
(AutoRes_IntegerColumn *self,
 int recordId,
 int value)
{
  int *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  
  dataArea = (int *)self->block->dataArea;
  assert(dataArea != NULL);

  dataArea[recordId] = value;
}

AutoRes_Table *AutoRes_IntegerColumn_table 
(AutoRes_IntegerColumn *self)
{
  return self->table;
}

void AutoRes_IntegerColumn_getName 
(AutoRes_IntegerColumn *self,
 char *name_OUT)
{
  strcpy (name_OUT, self->name);
}

int AutoRes_IntegerColumn_value 
(AutoRes_IntegerColumn *self,
 int recordId)
{
  int *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  
  dataArea = (int *)self->block->dataArea;
  assert(dataArea != NULL);

  return dataArea[recordId];
}

/* WARNING : test stage */
int *AutoRes_IntegerColumn_dataArea 
(AutoRes_IntegerColumn *self)
{
  assert(self->block != NULL);
  
  return (int *)self->block->dataArea;
}



/* Integer1DColumn */

AUTORES_IMPLEMENT_HANDLE(AutoRes_Integer1DColumn);

AutoRes_Integer1DColumn *AutoRes_Integer1DColumn_New 
(AutoRes_Table *table,
 const char *name,
 int size)
{
  AutoRes_Integer1DColumn *column;

  assert(0 < strlen (name));
  assert(strlen (name) < AUTORES_MAX_COLUMN_NAME_LENGTH);
  assert(0 < size);

  column = AutoRes_Integer1DColumn_NewImplementation ();

  column->table = table;
  column->size = size;
  column->block = NULL;

  column->name = (char *)malloc (strlen (name) + 1);
  assert(column->name != NULL);
  strcpy (column->name, name);

  return column;
}

void AutoRes_Integer1DColumn_delete 
(AutoRes_Integer1DColumn *self)
{
  free (self->name);
  self->name = NULL;

  if (self->block != NULL) {
    AutoRes_Memory_FreeBlock (self->block);
    self->block = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_Integer1DColumn_allocate 
(AutoRes_Integer1DColumn *self)
{
  int nReservedRecords;
  int memorySize;
  int flag;

  nReservedRecords = AutoRes_Table_nReservedRecords (self->table);
  memorySize = sizeof (int) * self->size * nReservedRecords;

  if (self->block != NULL) {
    flag = AutoRes_Memory_ReallocateBlock (self->block, memorySize);
    if (!flag) return 0;
  } else {
    self->block = AutoRes_Memory_AllocateBlock (memorySize);
    if (self->block == NULL) return 0;
  }
  return 1;
}

void AutoRes_Integer1DColumn_setLockFlag 
(AutoRes_Integer1DColumn *self,
 int flag)
{
  assert(self->block != NULL);

  AutoRes_Memory_SetLockFlagOnBlock (self->block, flag);
}

void AutoRes_Integer1DColumn_setValue 
(AutoRes_Integer1DColumn *self,
 int recordId,
 int index,
 int value)
{
  int *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  assert(0 <= index);
  assert(index < self->size);
  
  dataArea = (int *)self->block->dataArea;
  assert(dataArea != NULL);

  dataArea[self->size * recordId + index] = value;
}

AutoRes_Table *AutoRes_Integer1DColumn_table 
(AutoRes_Integer1DColumn *self)
{
  return self->table;
}

void AutoRes_Integer1DColumn_getName 
(AutoRes_Integer1DColumn *self,
 char *name_OUT)
{
  strcpy (name_OUT, self->name);
}

int AutoRes_Integer1DColumn_size 
(AutoRes_Integer1DColumn *self)
{
  return self->size;
}

int AutoRes_Integer1DColumn_value 
(AutoRes_Integer1DColumn *self,
 int recordId,
 int index)
{
  int *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  assert(0 <= index);
  assert(index < self->size);
  
  dataArea = (int *)self->block->dataArea;
  assert(dataArea != NULL);

  return dataArea[self->size * recordId + index];
}

/* WARNING : test stage */
int* AutoRes_Integer1DColumn_dataArea 
(AutoRes_Integer1DColumn *self)
{
  assert(self->block != NULL);
  
  return (int *)self->block->dataArea;
}






