/* autores_io.c : Input and Output */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_file.h"
#include "autores_field.h"
#include "autores_integer_field.h"
#include "autores_real_field.h"
#include "autores_io.h"



/* Input and Output */



/* InOut */

#define METADATA_KEYWORD_HEADER "HEADER" 
#define METADATA_KEYWORD_RECORD "RECORD" 
#define METADATA_KEYWORD_END "END" 

#define METADATA_KEYWORD_FIELD_INTEGER "INTEGER" 
#define METADATA_KEYWORD_FIELD_INTEGER1D "INTEGER1D" 
#define METADATA_KEYWORD_FIELD_REAL "REAL" 
#define METADATA_KEYWORD_FIELD_REAL1D "REAL1D" 

AUTORES_IMPLEMENT_HANDLE(AutoRes_InOut);

static void ClearFields 
(AutoRes_InOut *self)
{
  int iField;

  for (iField = 0; iField < self->nHeaderFields; iField++) {
    AutoRes_Field_delete 
      (self->headerFields[iField],
       self->headerFieldTypes[iField]);
    self->headerFields[iField] = NULL;
  }
  self->nHeaderFields = 0;

  for (iField = 0; iField < self->nRecordFields; iField++) {
    AutoRes_Field_delete 
      (self->recordFields[iField],
       self->recordFieldTypes[iField]);
    self->recordFields[iField] = NULL;
  }
  self->nRecordFields = 0;
}

AutoRes_InOut *AutoRes_InOut_New (void)
{
  AutoRes_InOut *inOut;

  inOut = AutoRes_InOut_NewImplementation ();

  inOut->nHeaderFields = 0;
  inOut->nRecordFields = 0;
  inOut->file = NULL;
  inOut->nRecords = 0;
  inOut->nRecordsPtr = NULL;
  inOut->recordIdFlag = 0;
  
  assert(inOut->file == NULL);

  return inOut;
}

void AutoRes_InOut_delete 
(AutoRes_InOut *self)
{
  assert(self->file == NULL);

  ClearFields (self);

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

void AutoRes_InOut_addFieldIntoHeader
(AutoRes_InOut *self,
 void *field,
 AutoRes_FieldType fieldType)
{
  assert(self->file == NULL);
  assert(self->nHeaderFields < AUTORES_MAX_IN_OUT_FIELDS);
  
  self->headerFieldTypes[self->nHeaderFields] = fieldType;
  self->headerFields[self->nHeaderFields] = field;
  self->nHeaderFields++;
}

void AutoRes_InOut_addIntegerFieldIntoHeader
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoHeader
    (self, field, AUTORES_FIELD_INTEGER);
}

void AutoRes_InOut_addInteger1DFieldIntoHeader
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoHeader
    (self, field, AUTORES_FIELD_INTEGER1D);
}

void AutoRes_InOut_addRealFieldIntoHeader
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoHeader
    (self, field, AUTORES_FIELD_REAL);
}

void AutoRes_InOut_addReal1DFieldIntoHeader
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoHeader
    (self, field, AUTORES_FIELD_REAL1D);
}

void AutoRes_InOut_addFieldIntoRecord
(AutoRes_InOut *self,
 void *field,
 AutoRes_FieldType fieldType)
{
  assert(self->file == NULL);
  assert(self->nRecordFields < AUTORES_MAX_IN_OUT_FIELDS);
  
  self->recordFieldTypes[self->nRecordFields] = fieldType;
  self->recordFields[self->nRecordFields] = field;
  self->nRecordFields++;
}

void AutoRes_InOut_addIntegerFieldIntoRecord
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoRecord
    (self, field, AUTORES_FIELD_INTEGER);
}

void AutoRes_InOut_addInteger1DFieldIntoRecord
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoRecord
    (self, field, AUTORES_FIELD_INTEGER1D);
}

void AutoRes_InOut_addRealFieldIntoRecord
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoRecord
    (self, field, AUTORES_FIELD_REAL);
}

void AutoRes_InOut_addReal1DFieldIntoRecord
(AutoRes_InOut *self,
 void *field)
{
  AutoRes_InOut_addFieldIntoRecord
    (self, field, AUTORES_FIELD_REAL1D);
}

void AutoRes_InOut_setRecordIdFlag
(AutoRes_InOut *self,
 int flag)
{
  self->recordIdFlag = flag;
}

static int ReadFieldMetaData 
(FILE *fp, const char *keyword,
 void **field_OUT, AutoRes_FieldType *fieldType_OUT)
{
  char fieldName[AUTORES_MAX_FIELD_NAME_LENGTH];
  
  if (strcmp (keyword, METADATA_KEYWORD_FIELD_INTEGER) == 0) {
    fscanf (fp, "%s", 
	    fieldName);

    *field_OUT = AutoRes_IntegerField_New (fieldName);
    *fieldType_OUT = AUTORES_FIELD_INTEGER;

  } else if (strcmp (keyword, METADATA_KEYWORD_FIELD_INTEGER1D) == 0) {
    int size;
    
    fscanf (fp, "%s %d", 
	    fieldName, &size);

    *field_OUT = AutoRes_Integer1DField_New (fieldName, size);
    *fieldType_OUT = AUTORES_FIELD_INTEGER1D;

  } else if (strcmp (keyword, METADATA_KEYWORD_FIELD_REAL) == 0) {
    fscanf (fp, "%s", 
	    fieldName);

    *field_OUT = AutoRes_RealField_New (fieldName);
    *fieldType_OUT = AUTORES_FIELD_REAL;

  } else if (strcmp (keyword, METADATA_KEYWORD_FIELD_REAL1D) == 0) {
    int size;
    
    fscanf (fp, "%s %d", 
	    fieldName, &size);

    *field_OUT = AutoRes_Real1DField_New (fieldName, size);
    *fieldType_OUT = AUTORES_FIELD_REAL1D;

  } else {
    assert(0);
    return 0;
  }

  return 1;
}

int /* boolean */
AutoRes_InOut_readMetaData
(AutoRes_InOut *self,
 const char *fileName)
{
  FILE *fp;
  char keyword[1024];
  char buf[1024];
  int flag;
  
  assert(0 < strlen (fileName));
  
  ClearFields (self);

  fp = fopen (fileName, "r");
  if (fp == NULL) {
    sprintf (buf, " read meta-data : file %s not found ",
	     fileName);
    AutoRes_Warning (buf);
    return 0;
  }

  fscanf (fp, "%s", 
	  keyword);
  if (!strcmp (keyword, METADATA_KEYWORD_HEADER) == 0) {
    sprintf (buf, " read meta-data : keyword must be %s ",
	     METADATA_KEYWORD_HEADER);
    AutoRes_Warning (buf);
    fclose (fp);
    return 0;
  }
  
  for (;;) {
    void *field;
    AutoRes_FieldType fieldType;

    fscanf (fp, "%s", 
	    keyword);
    if (strcmp (keyword, METADATA_KEYWORD_END) == 0) {
      break;
    }

    flag = ReadFieldMetaData (fp, keyword,
			      &field, &fieldType);
    if (!flag) {
      AutoRes_Warning (" read meta-data : illegal header field ");
      fclose (fp);
      return 0;
    }
    AutoRes_InOut_addFieldIntoHeader (self, field, fieldType);
  }

  fscanf (fp, "%s", 
	  keyword);
  if (!strcmp (keyword, METADATA_KEYWORD_RECORD) == 0) {
    sprintf (buf, " read meta-data : keyword must be %s ",
	     METADATA_KEYWORD_RECORD);
    AutoRes_Warning (buf);
    fclose (fp);
    return 0;
  }
  
  for (;;) {
    void *field;
    AutoRes_FieldType fieldType;

    fscanf (fp, "%s", 
	    keyword);
    if (strcmp (keyword, METADATA_KEYWORD_END) == 0) {
      break;
    }

    flag = ReadFieldMetaData (fp, keyword,
			      &field, &fieldType);
    if (!flag) {
      AutoRes_Warning (" read meta-data : illegal record field ");
      fclose (fp);
      return 0;
    }
    AutoRes_InOut_addFieldIntoRecord (self, field, fieldType);
  }

  fclose (fp);

  return 1;
}

static void WriteFieldMetaData 
(FILE *fp, void *field, AutoRes_FieldType fieldType)
{
  char fieldName[AUTORES_MAX_FIELD_NAME_LENGTH];
  
  AutoRes_Field_getName (field, fieldType,
			 fieldName);
  fprintf (fp, "  ");
  switch (fieldType) {
  case AUTORES_FIELD_INTEGER:
    fprintf (fp, "%s %s",
	     METADATA_KEYWORD_FIELD_INTEGER,
	     fieldName);
    break;
  case AUTORES_FIELD_INTEGER1D:
    {
      AutoRes_Integer1DField *field2 = (AutoRes_Integer1DField *)field;

      fprintf (fp, "%s %s %d",
	       METADATA_KEYWORD_FIELD_INTEGER1D,
	       fieldName, 
	       AutoRes_Integer1DField_size (field2));
    }
    break;
  case AUTORES_FIELD_REAL:
    fprintf (fp, "%s %s",
	     METADATA_KEYWORD_FIELD_REAL,
	     fieldName);
    break;
  case AUTORES_FIELD_REAL1D:
    {
      AutoRes_Real1DField *field2 = (AutoRes_Real1DField *)field;

      fprintf (fp, "%s %s %d",
	       METADATA_KEYWORD_FIELD_REAL1D,
	       fieldName, 
	       AutoRes_Real1DField_size (field2));
    }
    break;
  default:
    assert(0);
    break;
  }
  fprintf (fp, "\n");
}

int /* boolean */
AutoRes_InOut_writeMetaData
(AutoRes_InOut *self,
 const char *fileName)
{
  FILE *fp;
  int iField;

  assert(0 < strlen (fileName));
  
  fp = fopen (fileName, "w");
  if (fp == NULL) {
    return 0;
  }

  fprintf (fp, "%s\n",
	   METADATA_KEYWORD_HEADER);
  for (iField = 0; iField < self->nHeaderFields; iField++) {
    WriteFieldMetaData (fp, 
			self->headerFields[iField], 
			self->headerFieldTypes[iField]);
  }
  fprintf (fp, "%s\n",
	   METADATA_KEYWORD_END);

  fprintf (fp, "%s\n",
	   METADATA_KEYWORD_RECORD);
  for (iField = 0; iField < self->nRecordFields; iField++) {
    WriteFieldMetaData (fp, 
			self->recordFields[iField], 
			self->recordFieldTypes[iField]);
  }
  fprintf (fp, "%s\n",
	   METADATA_KEYWORD_END);

  fclose (fp);

  return 1;
}

void AutoRes_InOut_associateNRecords
(AutoRes_InOut *self,
 int *nRecordsPtr)
{
  assert(nRecordsPtr != NULL
	 || nRecordsPtr == NULL);
  
  self->nRecordsPtr = nRecordsPtr;
}

void AutoRes_InOut_setNRecords
(AutoRes_InOut *self,
 int nRecords)
{
  assert(0 <= nRecords);
  
  if (self->nRecordsPtr != NULL) {
    *self->nRecordsPtr = nRecords;
  } else {
    self->nRecords = nRecords;
  }
}

int /* boolean */
AutoRes_InOut_open 
(AutoRes_InOut *self,
 const char *fileName,
 int writeOpenFlag,
 int binaryFlag)
{
  int flag;
  
  assert(self->file == NULL);

  self->file = AutoRes_File_New ();
  assert(self->file != NULL);

  flag = AutoRes_File_open 
    (self->file, fileName, writeOpenFlag, binaryFlag);
  if (!flag) return 0;

  assert(self->file != NULL);

  return 1;
}

int /* boolean */
AutoRes_InOut_close 
(AutoRes_InOut *self)
{
  int flag;

  assert(self->file != NULL);

  flag = AutoRes_File_close (self->file);
  if (!flag) return 0;

  AutoRes_File_delete (self->file);
  self->file = NULL;

  assert(self->file == NULL);

  return 1;
}

static int /* boolean */
DoTopOfHeaderSection 
(AutoRes_InOut *self)
{
  int flag;

  if (self->nRecordsPtr != NULL) {
    flag = AutoRes_File_integer (self->file, self->nRecordsPtr);
    if (!flag) return 0;
  } else {
    flag = AutoRes_File_integer (self->file, &self->nRecords);
    if (!flag) return 0;
  }
  
  flag = AutoRes_File_nextLine (self->file);
  if (!flag) return 0;

  return 1;
}

static int /* boolean */
DoRecordIdField 
(AutoRes_File *file,
 int recordId)
{
  int writeFlag = AutoRes_File_writeFlag (file);
  int value;
  int flag;
      
  if (writeFlag) {
    value = recordId;
  }
  
  flag = AutoRes_File_integer (file, &value);
  if (!flag) return 0;
  
  if (!writeFlag) {
    assert(value == recordId);

    if (value != recordId) {
      AutoRes_Warning (" must be recordId ");
      return 0;
    }
  }

  return 1;
}

static int /* boolean */
DoField 
(void *field, 
 AutoRes_FieldType fieldType,
 AutoRes_File *file)
{
  int writeFlag = AutoRes_File_writeFlag (file);
  int flag;
  
  switch (fieldType) {
  case AUTORES_FIELD_INTEGER:
    {
      AutoRes_IntegerField *field2 = (AutoRes_IntegerField *)field;
      int value;
      
      if (writeFlag) {
	value = AutoRes_IntegerField_value (field2);
      }
      
      flag = AutoRes_File_integer (file, &value);
      if (!flag) return 0;
      
      if (!writeFlag) {
	AutoRes_IntegerField_setValue (field2, value);
      }
    }
    break;
  case AUTORES_FIELD_INTEGER1D:
    {
      AutoRes_Integer1DField *field2 = (AutoRes_Integer1DField *)field;
      int size = AutoRes_Integer1DField_size (field2);
      int index;
      
      for (index = 0; index < size; index++) {
	int value;

	if (writeFlag) {
	  value = AutoRes_Integer1DField_value (field2, index);
	}
	
	flag = AutoRes_File_integer (file, &value);
	if (!flag) return 0;
	
	if (!writeFlag) {
	  AutoRes_Integer1DField_setValue (field2, index, value);
	}
      }
    }
    break;
  case AUTORES_FIELD_REAL:
    {
      AutoRes_RealField *field2 = (AutoRes_RealField *)field;
      double value;
      
      if (writeFlag) {
	value = AutoRes_RealField_value (field2);
      }
      
      flag = AutoRes_File_real (file, &value);
      if (!flag) return 0;
      
      if (!writeFlag) {
	AutoRes_RealField_setValue (field2, value);
      }
    }
    break;
  case AUTORES_FIELD_REAL1D:
    {
      AutoRes_Real1DField *field2 = (AutoRes_Real1DField *)field;
      int size = AutoRes_Real1DField_size (field2);
      int index;
      
      for (index = 0; index < size; index++) {
	double value;

	if (writeFlag) {
	  value = AutoRes_Real1DField_value (field2, index);
	}
	
	flag = AutoRes_File_real (file, &value);
	if (!flag) return 0;
	
	if (!writeFlag) {
	  AutoRes_Real1DField_setValue (field2, index, value);
	}
      }
    }
    break;
  default:
    assert(0);
    break;
  }

  return 1;
}

static int /* boolean */
DoOptionalHeaderSection 
(AutoRes_InOut *self)
{
  int iField;
  int flag;

  for (iField = 0; iField < self->nHeaderFields; iField++) {
    flag = DoField (self->headerFields[iField],
		    self->headerFieldTypes[iField],
		    self->file);
    if (!flag) return 0;
  }

  if (0 < self->nHeaderFields) {
    flag = AutoRes_File_nextLine (self->file);
    if (!flag) return 0;
  }

  return 1;
}

int /* boolean */
AutoRes_InOut_doHeaderSection 
(AutoRes_InOut *self)
{
  int flag;
  
  assert(self->file != NULL);

  flag = DoTopOfHeaderSection (self);
  if (!flag) return 0;
  
  flag = DoOptionalHeaderSection (self);
  if (!flag) return 0;
  
  return 1;
}

int /* boolean */
AutoRes_InOut_doEachOfRecordSection 
(AutoRes_InOut *self,
 int recordId)
{
  int iField;
  int flag;

  assert(self->file != NULL);
  assert(0 <= recordId);

  if (self->recordIdFlag) {
    flag = DoRecordIdField (self->file, recordId);
    if (!flag) return 0;
  }

  for (iField = 0; iField < self->nRecordFields; iField++) {
    flag = DoField (self->recordFields[iField],
		    self->recordFieldTypes[iField],
		    self->file);
    if (!flag) return 0;
  }
  
  if (0 < self->nRecordFields) {
    flag = AutoRes_File_nextLine (self->file);
    if (!flag) return 0;
  }

  return 1;
}

int AutoRes_InOut_nHeaderFields 
(AutoRes_InOut *self)
{
  return self->nHeaderFields;
}

void *AutoRes_InOut_headerField 
(AutoRes_InOut *self,
 int fieldId)
{
  assert(0 <= fieldId);
  assert(fieldId < self->nHeaderFields);
  
  return self->headerFields[fieldId];
}

AutoRes_FieldType AutoRes_InOut_headerFieldType
(AutoRes_InOut *self,
 int fieldId)
{
  assert(0 <= fieldId);
  assert(fieldId < self->nHeaderFields);
  
  return self->headerFieldTypes[fieldId];
}

int AutoRes_InOut_headerFieldIdOfName
(AutoRes_InOut *self,
 const char* fieldName)
{
  int iField;
  
  assert(0 < strlen (fieldName));
  assert(strlen (fieldName) < AUTORES_MAX_FIELD_NAME_LENGTH);
  
  for (iField = 0; iField < self->nHeaderFields; iField++) {
    char name[AUTORES_MAX_FIELD_NAME_LENGTH];

    AutoRes_Field_getName (self->headerFields[iField],
			   self->headerFieldTypes[iField],
			   name);
    if (strcmp (name, fieldName) == 0) {
      return iField;
    }
  }
  return -1;
}

void *AutoRes_InOut_headerFieldOfName 
(AutoRes_InOut *self,
 const char* fieldName)
{
  int fieldId = AutoRes_InOut_headerFieldIdOfName (self, fieldName);

  if (fieldId == -1) {
    return NULL;
  } else {
    return AutoRes_InOut_headerField (self, fieldId);
  }
}

int AutoRes_InOut_nRecordFields 
(AutoRes_InOut *self)
{
  return self->nRecordFields;
}

void *AutoRes_InOut_recordField 
(AutoRes_InOut *self,
 int fieldId)
{
  assert(0 <= fieldId);
  assert(fieldId < self->nRecordFields);
  
  return self->recordFields[fieldId];
}

AutoRes_FieldType AutoRes_InOut_recordFieldType
(AutoRes_InOut *self,
 int fieldId)
{
  assert(0 <= fieldId);
  assert(fieldId < self->nRecordFields);
  
  return self->recordFieldTypes[fieldId];
}

int AutoRes_InOut_recordFieldIdOfName
(AutoRes_InOut *self,
 const char* fieldName)
{
  int iField;
  
  assert(0 < strlen (fieldName));
  assert(strlen (fieldName) < AUTORES_MAX_FIELD_NAME_LENGTH);
  
  for (iField = 0; iField < self->nRecordFields; iField++) {
    char name[AUTORES_MAX_FIELD_NAME_LENGTH];

    AutoRes_Field_getName (self->recordFields[iField],
			   self->recordFieldTypes[iField],
			   name);
    if (strcmp (name, fieldName) == 0) {
      return iField;
    }
  }
  return -1;
}

void *AutoRes_InOut_recordFieldOfName 
(AutoRes_InOut *self,
 const char* fieldName)
{
  int fieldId = AutoRes_InOut_recordFieldIdOfName (self, fieldName);

  if (fieldId == -1) {
    return NULL;
  } else {
    return AutoRes_InOut_recordField (self, fieldId);
  }
}

struct _AutoRes_File *AutoRes_InOut_file 
(AutoRes_InOut *self)
{
  return self->file;
}

int AutoRes_InOut_nRecords 
(AutoRes_InOut *self)
{
  if (self->nRecordsPtr != NULL) {
    return *self->nRecordsPtr;
  } else {
    return self->nRecords;
  }
}

int /* boolean */ 
AutoRes_InOut_recordIdFlag 
(AutoRes_InOut *self)
{
  return self->recordIdFlag;
}
