/* autores_list.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_list.h"



/* List */

#define TABLE_NAME_LINKS "AutoRes:links"
#define COLUMN_NAME_NEXT_LINK_ID "AutoRes:nextLinkId"
#define COLUMN_NAME_PREV_LINK_ID "AutoRes:prevLinkId"
#define COLUMN_NAME_OBJECT_ID "AutoRes:objectId"

#define NULL_LINK_ID -1
#define NULL_OBJECT_ID -1

AUTORES_IMPLEMENT_HANDLE(AutoRes_List);

AutoRes_List *AutoRes_List_New (void)
{
  AutoRes_List *list;

  list = AutoRes_List_NewImplementation ();

  list->startLinkId = NULL_LINK_ID;
  list->endLinkId = NULL_LINK_ID;
  list->freeLinkId = NULL_LINK_ID;

  list->links = NULL;
  list->links_nextLinkId = NULL;
  list->links_prevLinkId = NULL;
  list->links_objectId = NULL;

  return list;
}

void AutoRes_List_delete 
(AutoRes_List *self)
{
  if (self->links != NULL) {
    AutoRes_Table_delete (self->links);
    self->links = NULL;
    self->links_nextLinkId = NULL;
    self->links_prevLinkId = NULL;
    self->links_objectId = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_List_allocate 
(AutoRes_List *self,
 int nInitialLinks)
{
  int flag;

  assert(self->links == NULL);
  assert(0 < nInitialLinks);

  self->links = AutoRes_Table_New (TABLE_NAME_LINKS);
  assert(self->links != NULL);
  AutoRes_Table_reserveRecords (self->links, nInitialLinks);

  self->links_nextLinkId = AutoRes_IntegerColumn_New 
    (self->links, COLUMN_NAME_NEXT_LINK_ID);
  assert(self->links_nextLinkId != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->links, self->links_nextLinkId);
  if (!flag) return 0;

  self->links_prevLinkId = AutoRes_IntegerColumn_New 
    (self->links, COLUMN_NAME_PREV_LINK_ID);
  assert(self->links_prevLinkId != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->links, self->links_prevLinkId);
  if (!flag) return 0;

  self->links_objectId = AutoRes_IntegerColumn_New 
    (self->links, COLUMN_NAME_OBJECT_ID);
  assert(self->links_objectId != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->links, self->links_objectId);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoRes_List_addLinkAtFirst
(AutoRes_List *self,
 int objectId)
{
  int newLinkId;

  assert(self->links != NULL);
  assert(0 <= objectId);
  
  if (0 <= self->freeLinkId) {
    newLinkId = self->freeLinkId;
    self->freeLinkId = AUTORES_I(self->links_nextLinkId, newLinkId);
  } else {
    newLinkId = AutoRes_Table_addRecord (self->links);
    if (newLinkId == -1) return 0;
  }

  if (self->startLinkId == NULL_LINK_ID) {
    assert(self->endLinkId == NULL_LINK_ID);

    self->startLinkId = newLinkId;
    self->endLinkId = newLinkId;
    AUTORES_I_SET(self->links_nextLinkId, newLinkId, NULL_LINK_ID);
    AUTORES_I_SET(self->links_prevLinkId, newLinkId, NULL_LINK_ID);

  } else {
    AUTORES_I_SET(self->links_prevLinkId, self->startLinkId, newLinkId);
    AUTORES_I_SET(self->links_prevLinkId, newLinkId, NULL_LINK_ID);
    AUTORES_I_SET(self->links_nextLinkId, newLinkId, self->startLinkId);
    self->startLinkId = newLinkId;
  }

  AUTORES_I_SET(self->links_objectId, newLinkId, objectId);

  return 1;
}

int /* boolean */
AutoRes_List_addLinkAtLast 
(AutoRes_List *self,
 int objectId)
{
  int newLinkId;

  assert(self->links != NULL);
  assert(0 <= objectId);
  
  if (0 <= self->freeLinkId) {
    newLinkId = self->freeLinkId;
    self->freeLinkId = AUTORES_I(self->links_nextLinkId, newLinkId);
  } else {
    newLinkId = AutoRes_Table_addRecord (self->links);
    if (newLinkId == -1) return 0;
  }

  if (self->endLinkId == NULL_LINK_ID) {
    assert(self->startLinkId == NULL_LINK_ID);

    self->startLinkId = newLinkId;
    self->endLinkId = newLinkId;
    AUTORES_I_SET(self->links_nextLinkId, newLinkId, NULL_LINK_ID);
    AUTORES_I_SET(self->links_prevLinkId, newLinkId, NULL_LINK_ID);

  } else {
    AUTORES_I_SET(self->links_nextLinkId, self->endLinkId, newLinkId);
    AUTORES_I_SET(self->links_nextLinkId, newLinkId, NULL_LINK_ID);
    AUTORES_I_SET(self->links_prevLinkId, newLinkId, self->endLinkId);
    self->endLinkId = newLinkId;
  }

  AUTORES_I_SET(self->links_objectId, newLinkId, objectId);

  return 1;
}

void AutoRes_List_removeLink 
(AutoRes_List *self,
 int linkId)
{
  int nextLinkId;
  int prevLinkId;

  assert(self->links != NULL);
  
  nextLinkId = AUTORES_I(self->links_nextLinkId, linkId);
  prevLinkId = AUTORES_I(self->links_prevLinkId, linkId);

  if (nextLinkId == NULL_LINK_ID) {
    assert(self->endLinkId == linkId);
    self->endLinkId = prevLinkId;
  } else {
    AUTORES_I_SET(self->links_prevLinkId, nextLinkId, prevLinkId);
  }

  if (prevLinkId == NULL_LINK_ID) {
    assert(self->startLinkId == linkId);
    self->startLinkId = nextLinkId;
  } else {
    AUTORES_I_SET(self->links_nextLinkId, prevLinkId, nextLinkId);
  }

  AUTORES_I_SET(self->links_nextLinkId, linkId, self->freeLinkId);
  AUTORES_I_SET(self->links_prevLinkId, linkId, NULL_LINK_ID);
  AUTORES_I_SET(self->links_objectId, linkId, NULL_OBJECT_ID);
  self->freeLinkId = linkId;
}

int AutoRes_List_startLinkId 
(AutoRes_List *self)
{
  assert(self->links != NULL);

  return self->startLinkId;
}

int AutoRes_List_nextLinkId 
(AutoRes_List *self,
 int linkId)
{
  assert(self->links != NULL);

  return AUTORES_I(self->links_nextLinkId, linkId);
}

int AutoRes_List_endLinkId 
(AutoRes_List *self)
{
  assert(self->links != NULL);

  return self->endLinkId;
}

int AutoRes_List_prevLinkId 
(AutoRes_List *self,
 int linkId)
{
  assert(self->links != NULL);

  return AUTORES_I(self->links_prevLinkId, linkId);
}

int AutoRes_List_objectId 
(AutoRes_List *self,
 int linkId)
{
  assert(self->links != NULL);

  return AUTORES_I(self->links_objectId, linkId);
}

