/* autores_list_table.c : table of multiple lists */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_list_table.h"



/* List Table */

#define TABLE_NAME_LISTS "AutoRes:lists"
#define COLUMN_NAME_START_LINK_ID "AutoRes:startLinkId"

#define TABLE_NAME_LINKS "AutoRes:links"
#define COLUMN_NAME_NEXT_LINK_ID "AutoRes:nextLinkId"
#define COLUMN_NAME_OBJECT_ID "AutoRes:objectId"

#define NULL_LINK_ID -1
#define NULL_OBJECT_ID -1

AUTORES_IMPLEMENT_HANDLE(AutoRes_ListTable);

AutoRes_ListTable *AutoRes_ListTable_New (void)
{
  AutoRes_ListTable *listTable;
  int flag;

  listTable = AutoRes_ListTable_NewImplementation ();

  listTable->nLists = 0;

  listTable->lists = AutoRes_Table_New (TABLE_NAME_LISTS);
  assert(listTable->lists != NULL);
  
  listTable->lists_startLinkId = AutoRes_IntegerColumn_New 
    (listTable->lists, COLUMN_NAME_START_LINK_ID);
  assert(listTable->lists_startLinkId != NULL);
  
  flag = AutoRes_Table_addIntegerColumn 
    (listTable->lists, listTable->lists_startLinkId);
  assert(flag);
  
  listTable->links = AutoRes_Table_New (TABLE_NAME_LINKS);
  assert(listTable->links != NULL);
  
  listTable->links_nextLinkId = AutoRes_IntegerColumn_New 
    (listTable->links, COLUMN_NAME_NEXT_LINK_ID);
  assert(listTable->links_nextLinkId != NULL);
  
  flag = AutoRes_Table_addIntegerColumn 
    (listTable->links, listTable->links_nextLinkId);
  assert(flag);

  listTable->links_objectId = AutoRes_IntegerColumn_New 
    (listTable->links, COLUMN_NAME_OBJECT_ID);
  assert(listTable->links_objectId != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (listTable->links, listTable->links_objectId);
  assert(flag);

  listTable->freeLinkId = NULL_LINK_ID;

  return listTable;
}

void AutoRes_ListTable_delete 
(AutoRes_ListTable *self)
{
  AutoRes_Table_delete (self->lists);
  self->lists = NULL;
  self->lists_startLinkId = NULL;

  AutoRes_Table_delete (self->links);
  self->links = NULL;
  self->links_nextLinkId = NULL;
  self->links_objectId = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_ListTable_reserve 
(AutoRes_ListTable *self,
 int nLists,
 int nAverageLinksPerList)
{
  int iList;
  int flag;

  assert(0 <= nLists);
  assert(0 < nAverageLinksPerList);
  
  self->nLists = nLists;
  
  flag = AutoRes_Table_allocateRecords (self->lists, nLists);
  if (!flag) return 0;
  
  for (iList = 0; iList < nLists; iList++) {
    AUTORES_I_SET(self->lists_startLinkId, iList, NULL_LINK_ID); 
  }
  
  flag = AutoRes_Table_reserveRecords 
    (self->links, nLists * nAverageLinksPerList);
  if (!flag) return 0;
  
  return 1;
}

static void AddLinkIntoFreeList 
(AutoRes_ListTable *self,
 int linkId)
{
  int startLinkId;

  startLinkId = self->freeLinkId;
  self->freeLinkId = linkId;
  AUTORES_I_SET(self->links_nextLinkId, linkId, startLinkId);
  AUTORES_I_SET(self->links_objectId, linkId, NULL_OBJECT_ID);
}

static int RemoveLinkFromFreeList 
(AutoRes_ListTable *self)
{
  int linkId;

  linkId = self->freeLinkId;
  if (linkId != NULL_LINK_ID) {
    self->freeLinkId = AUTORES_I(self->links_nextLinkId, linkId);
  }

  assert(linkId == NULL_LINK_ID 
	 || linkId != NULL_LINK_ID); 

  return linkId;
}

int AutoRes_ListTable_addObject 
(AutoRes_ListTable *self,
 int listId, 
 int objectId)
{
  int newLinkId;
  int startLinkId;

  assert(0 <= listId);
  assert(listId < self->nLists);
  assert(0 <= objectId);
  
  newLinkId = RemoveLinkFromFreeList (self);
  if (newLinkId == NULL_LINK_ID) {
    newLinkId = AutoRes_Table_addRecord (self->links);
    if (newLinkId == -1) return -1;
  }
  assert(newLinkId != NULL_LINK_ID);

  startLinkId = AUTORES_I(self->lists_startLinkId, listId);
  AUTORES_I_SET(self->lists_startLinkId, listId, newLinkId);
  AUTORES_I_SET(self->links_nextLinkId, newLinkId, startLinkId);
  AUTORES_I_SET(self->links_objectId, newLinkId, objectId);

  return newLinkId;
}

void AutoRes_ListTable_removeObject 
(AutoRes_ListTable *self,
 int listId, 
 int objectId)
{
  int currentLinkId;
  int nextLinkId;

  assert(0 <= listId);
  assert(listId < self->nLists);
  assert(0 <= objectId);
  
  currentLinkId = AUTORES_I(self->lists_startLinkId, listId);
  if (currentLinkId == NULL_LINK_ID) {
    return;
  }
  assert(currentLinkId != NULL_LINK_ID);

  if (AUTORES_I(self->links_objectId, currentLinkId) == objectId) {
    AUTORES_I_SET(self->lists_startLinkId, listId,
		  AUTORES_I(self->links_nextLinkId, currentLinkId));
    AddLinkIntoFreeList (self, currentLinkId);
    return;
  }

  for (;;) {
    assert(AUTORES_I(self->links_objectId, currentLinkId) != objectId);
    
    nextLinkId = AUTORES_I(self->links_nextLinkId, currentLinkId);
    if (nextLinkId == NULL_LINK_ID) {
      return;
    }
    assert(nextLinkId != NULL_LINK_ID);
    
    if (AUTORES_I(self->links_objectId, nextLinkId) == objectId) {
      AUTORES_I_SET(self->links_nextLinkId, currentLinkId,
		    AUTORES_I(self->links_nextLinkId, nextLinkId));
      AddLinkIntoFreeList (self, nextLinkId);
      return;
    }
    assert(AUTORES_I(self->links_objectId, nextLinkId) != objectId);

    currentLinkId = nextLinkId;
  }
}

int AutoRes_ListTable_nLists 
(AutoRes_ListTable *self)
{
  return self->nLists;
}

int AutoRes_ListTable_startLinkId 
(AutoRes_ListTable *self,
 int listId)
{
  assert(0 <= listId);
  assert(listId < self->nLists);

  return AUTORES_I(self->lists_startLinkId, listId);
}

int AutoRes_ListTable_nextLinkId 
(AutoRes_ListTable *self,
 int linkId)
{
  return AUTORES_I(self->links_nextLinkId, linkId);
}

int AutoRes_ListTable_objectId 
(AutoRes_ListTable *self,
 int linkId)
{
  return AUTORES_I(self->links_objectId, linkId);
}

