/* autores_ordered.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_ordered.h"



/* Ordered Collection */

#define COLUMN_NAME_NEXT_RECORD_ID "AutoRes:nextRecordId"
#define COLUMN_NAME_PREV_RECORD_ID "AutoRes:prevRecordId"

#define NULL_RECORD_ID -1
#define FREE_RECORD_ID -2

AUTORES_IMPLEMENT_HANDLE(AutoRes_OrderedCollection);

AutoRes_OrderedCollection *AutoRes_OrderedCollection_New 
(AutoRes_Table *table)
{
  AutoRes_OrderedCollection *collection;

  collection = AutoRes_OrderedCollection_NewImplementation ();

  collection->table = table;
  collection->startRecordId = NULL_RECORD_ID;
  collection->endRecordId = NULL_RECORD_ID;
  collection->freeRecordId = NULL_RECORD_ID;
  collection->nextRecordIdColumn = NULL;
  collection->prevRecordIdColumn = NULL;

  return collection;
}

void AutoRes_OrderedCollection_delete 
(AutoRes_OrderedCollection *self)
{
  if (self->nextRecordIdColumn != NULL) {
    AutoRes_Table_removeColumn 
      (self->table, 
       AutoRes_Table_columnIdOfColumn 
       (self->table, self->nextRecordIdColumn));
    self->nextRecordIdColumn = NULL;
  }

  if (self->prevRecordIdColumn != NULL) {
    AutoRes_Table_removeColumn 
      (self->table, 
       AutoRes_Table_columnIdOfColumn 
       (self->table, self->prevRecordIdColumn));
    self->prevRecordIdColumn = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_OrderedCollection_makeColumn 
(AutoRes_OrderedCollection *self)
{
  int flag;

  assert(self->nextRecordIdColumn == NULL);
  assert(self->prevRecordIdColumn == NULL);

  self->nextRecordIdColumn = AutoRes_IntegerColumn_New 
    (self->table, COLUMN_NAME_NEXT_RECORD_ID);
  assert(self->nextRecordIdColumn != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->table, self->nextRecordIdColumn);
  if (!flag) return 0;

  self->prevRecordIdColumn = AutoRes_IntegerColumn_New 
    (self->table, COLUMN_NAME_PREV_RECORD_ID);
  assert(self->prevRecordIdColumn != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->table, self->prevRecordIdColumn);
  if (!flag) return 0;

  assert(self->nextRecordIdColumn != NULL);
  assert(self->prevRecordIdColumn != NULL);

  return 1;
}

void AutoRes_OrderedCollection_initializeColumn 
(AutoRes_OrderedCollection *self)
{
  int nRecords = AutoRes_Table_nRecords (self->table);
  int iRecord;
  
  if (0 < nRecords) {
    self->startRecordId = 0;
    self->endRecordId = nRecords - 1;

    for (iRecord = 0; iRecord < nRecords; iRecord++) {
      if (iRecord == nRecords - 1) {
	AUTORES_I_SET(self->nextRecordIdColumn, iRecord, NULL_RECORD_ID);
      } else {
	AUTORES_I_SET(self->nextRecordIdColumn, iRecord, iRecord + 1);
      }

      if (iRecord == 0) {
	AUTORES_I_SET(self->prevRecordIdColumn, iRecord, NULL_RECORD_ID);
      } else {
	AUTORES_I_SET(self->prevRecordIdColumn, iRecord, iRecord - 1);
      }
    }

  } else {
    self->startRecordId = NULL_RECORD_ID;
    self->endRecordId = NULL_RECORD_ID;
  }

  self->freeRecordId = NULL_RECORD_ID;
}

int AutoRes_OrderedCollection_addRecord 
(AutoRes_OrderedCollection *self)
{
  int newRecordId;

  assert(0 < AutoRes_Table_nReservedRecords (self->table));
  assert(self->nextRecordIdColumn != NULL);
  assert(self->prevRecordIdColumn != NULL);

  if (self->freeRecordId != NULL_RECORD_ID) {
    newRecordId = self->freeRecordId;
    self->freeRecordId = AUTORES_I(self->nextRecordIdColumn, newRecordId);
  } else {
    newRecordId = AutoRes_Table_addRecord (self->table);
    if (newRecordId == -1) return -1;
  }

  if (self->endRecordId == NULL_RECORD_ID) {
    assert(self->startRecordId == NULL_RECORD_ID);

    self->startRecordId = newRecordId;
    self->endRecordId = newRecordId;
    AUTORES_I_SET(self->nextRecordIdColumn, newRecordId, NULL_RECORD_ID);
    AUTORES_I_SET(self->prevRecordIdColumn, newRecordId, NULL_RECORD_ID);

  } else {
    /* add the new record after */

    AUTORES_I_SET(self->nextRecordIdColumn, self->endRecordId, newRecordId);
    AUTORES_I_SET(self->nextRecordIdColumn, newRecordId, NULL_RECORD_ID);
    AUTORES_I_SET(self->prevRecordIdColumn, newRecordId, self->endRecordId);
    self->endRecordId = newRecordId;
  }

  return newRecordId;
}

void AutoRes_OrderedCollection_removeRecord 
(AutoRes_OrderedCollection *self,
 int recordId)
{
  int nextRecordId;
  int prevRecordId;

  assert(self->nextRecordIdColumn != NULL);
  assert(self->prevRecordIdColumn != NULL);

  nextRecordId = AUTORES_I(self->nextRecordIdColumn, recordId);
  prevRecordId = AUTORES_I(self->prevRecordIdColumn, recordId);
  assert(prevRecordId != FREE_RECORD_ID);

  if (nextRecordId == NULL_RECORD_ID) {
    assert(self->endRecordId == recordId);
    self->endRecordId = prevRecordId;
  } else {
    AUTORES_I_SET(self->prevRecordIdColumn, nextRecordId, prevRecordId);
  }

  if (prevRecordId == NULL_RECORD_ID) {
    assert(self->startRecordId == recordId);
    self->startRecordId = nextRecordId;
  } else {
    AUTORES_I_SET(self->nextRecordIdColumn, prevRecordId, nextRecordId);
  }

  AUTORES_I_SET(self->nextRecordIdColumn, recordId, self->freeRecordId);
  AUTORES_I_SET(self->prevRecordIdColumn, recordId, FREE_RECORD_ID);
  self->freeRecordId = recordId;
}

AutoRes_Table *AutoRes_OrderedCollection_table 
(AutoRes_OrderedCollection *self)
{
  return self->table;
}

int /* boolean */
AutoRes_OrderedCollection_recordIsUsed 
(AutoRes_OrderedCollection *self,
 int recordId)
{
  return AUTORES_I(self->prevRecordIdColumn, recordId) != FREE_RECORD_ID;
}

int AutoRes_OrderedCollection_startRecordId 
(AutoRes_OrderedCollection *self)
{
  return self->startRecordId;
}

int AutoRes_OrderedCollection_endRecordId 
(AutoRes_OrderedCollection *self)
{
  return self->endRecordId;
}

int AutoRes_OrderedCollection_nextRecordId 
(AutoRes_OrderedCollection *self,
 int recordId)
{
  return AUTORES_I(self->nextRecordIdColumn, recordId);
}

int AutoRes_OrderedCollection_prevRecordId 
(AutoRes_OrderedCollection *self,
 int recordId)
{
  return AUTORES_I(self->prevRecordIdColumn, recordId);
}



