/* autores_real_column.c : Table-based Memory Manager : Real Column */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_real_column.h"



/* Table-based Memory Manager : Real Column */



/* RealColumn */

AUTORES_IMPLEMENT_HANDLE(AutoRes_RealColumn);

AutoRes_RealColumn *AutoRes_RealColumn_New 
(AutoRes_Table *table,
 const char *name)
{
  AutoRes_RealColumn *column;

  assert(0 < strlen (name));
  assert(strlen (name) < AUTORES_MAX_COLUMN_NAME_LENGTH);

  column = AutoRes_RealColumn_NewImplementation ();

  column->table = table;
  column->block = NULL;

  column->name = (char *)malloc (strlen (name) + 1);
  assert(column->name != NULL);
  strcpy (column->name, name);

  return column;
}

void AutoRes_RealColumn_delete 
(AutoRes_RealColumn *self)
{
  free (self->name);
  self->name = NULL;

  if (self->block != NULL) {
    AutoRes_Memory_FreeBlock (self->block);
    self->block = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_RealColumn_allocate 
(AutoRes_RealColumn *self)
{
  int nReservedRecords;
  int memorySize;
  int flag;

  nReservedRecords = AutoRes_Table_nReservedRecords (self->table);
  memorySize = sizeof (double) * nReservedRecords;

  if (self->block != NULL) {
    flag = AutoRes_Memory_ReallocateBlock (self->block, memorySize);
    if (!flag) return 0;
  } else {
    self->block = AutoRes_Memory_AllocateBlock (memorySize);
    if (self->block == NULL) return 0;
  }
  return 1;
}

void AutoRes_RealColumn_setLockFlag 
(AutoRes_RealColumn *self,
 int flag)
{
  assert(self->block != NULL);

  AutoRes_Memory_SetLockFlagOnBlock (self->block, flag);
}

void AutoRes_RealColumn_setValue 
(AutoRes_RealColumn *self,
 int recordId,
 double value)
{
  double *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  
  dataArea = (double *)self->block->dataArea;
  assert(dataArea != NULL);

  dataArea[recordId] = value;
}

AutoRes_Table *AutoRes_RealColumn_table 
(AutoRes_RealColumn *self)
{
  return self->table;
}

void AutoRes_RealColumn_getName 
(AutoRes_RealColumn *self,
 char *name_OUT)
{
  strcpy (name_OUT, self->name);
}

double AutoRes_RealColumn_value 
(AutoRes_RealColumn *self,
 int recordId)
{
  double *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  
  dataArea = (double *)self->block->dataArea;
  assert(dataArea != NULL);

  return dataArea[recordId];
}

/* WARNING : test stage */
double *AutoRes_RealColumn_dataArea 
(AutoRes_RealColumn *self)
{
  assert(self->block != NULL);
  
  return (double *)self->block->dataArea;
}



/* Real1DColumn */

AUTORES_IMPLEMENT_HANDLE(AutoRes_Real1DColumn);

AutoRes_Real1DColumn *AutoRes_Real1DColumn_New 
(AutoRes_Table *table,
 const char *name,
 int size)
{
  AutoRes_Real1DColumn *column;

  assert(0 < strlen (name));
  assert(strlen (name) < AUTORES_MAX_COLUMN_NAME_LENGTH);
  assert(0 < size);

  column = AutoRes_Real1DColumn_NewImplementation ();

  column->table = table;
  column->size = size;
  column->block = NULL;

  column->name = (char *)malloc (strlen (name) + 1);
  assert(column->name != NULL);
  strcpy (column->name, name);

  return column;
}

void AutoRes_Real1DColumn_delete 
(AutoRes_Real1DColumn *self)
{
  free (self->name);
  self->name = NULL;

  if (self->block != NULL) {
    AutoRes_Memory_FreeBlock (self->block);
    self->block = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_Real1DColumn_allocate 
(AutoRes_Real1DColumn *self)
{
  int nReservedRecords;
  int memorySize;
  int flag;

  nReservedRecords = AutoRes_Table_nReservedRecords (self->table);
  memorySize = sizeof (double) * self->size * nReservedRecords;

  if (self->block != NULL) {
    flag = AutoRes_Memory_ReallocateBlock (self->block, memorySize);
    if (!flag) return 0;
  } else {
    self->block = AutoRes_Memory_AllocateBlock (memorySize);
    if (self->block == NULL) return 0;
  }
  return 1;
}

void AutoRes_Real1DColumn_setLockFlag 
(AutoRes_Real1DColumn *self,
 int flag)
{
  assert(self->block != NULL);

  AutoRes_Memory_SetLockFlagOnBlock (self->block, flag);
}

void AutoRes_Real1DColumn_setValue 
(AutoRes_Real1DColumn *self,
 int recordId,
 int index,
 double value)
{
  double *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  assert(0 <= index);
  assert(index < self->size);
  
  dataArea = (double *)self->block->dataArea;
  assert(dataArea != NULL);

  dataArea[self->size * recordId + index] = value;
}

AutoRes_Table *AutoRes_Real1DColumn_table 
(AutoRes_Real1DColumn *self)
{
  return self->table;
}

void AutoRes_Real1DColumn_getName 
(AutoRes_Real1DColumn *self,
 char *name_OUT)
{
  strcpy (name_OUT, self->name);
}

int AutoRes_Real1DColumn_size 
(AutoRes_Real1DColumn *self)
{
  return self->size;
}

double AutoRes_Real1DColumn_value 
(AutoRes_Real1DColumn *self,
 int recordId,
 int index)
{
  double *dataArea;

  assert(self->block != NULL);
  assert(AutoRes_Table_recordIdIsValid (self->table, recordId));
  assert(0 <= index);
  assert(index < self->size);
  
  dataArea = (double *)self->block->dataArea;
  assert(dataArea != NULL);

  return dataArea[self->size * recordId + index];
}

/* WARNING : test stage */
double* AutoRes_Real1DColumn_dataArea 
(AutoRes_Real1DColumn *self)
{
  assert(self->block != NULL);
  
  return (double *)self->block->dataArea;
}



