/* autores_table_io.c : Table I/O */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_file.h"
#include "autores_field.h"
#include "autores_integer_field.h"
#include "autores_real_field.h"
#include "autores_io.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_table_io.h"



/* Table I/O */



/* TableInOut */

AUTORES_IMPLEMENT_HANDLE(AutoRes_TableInOut);

AutoRes_TableInOut *AutoRes_TableInOut_New 
(AutoRes_Table *table)
{
  AutoRes_TableInOut *tableInOut;

  tableInOut = AutoRes_TableInOut_NewImplementation ();

  tableInOut->table = table;
  tableInOut->nColumns = 0;
  tableInOut->inOut = AutoRes_InOut_New ();
  
  return tableInOut;
}

void AutoRes_TableInOut_delete 
(AutoRes_TableInOut *self)
{
  AutoRes_InOut_delete (self->inOut);
  self->inOut = NULL;

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

void AutoRes_TableInOut_associateColumnWithField 
(AutoRes_TableInOut *self,
 void *column,
 AutoRes_ColumnType columnType,
 void *field,
 AutoRes_FieldType fieldType)
{
  switch (columnType) {
  case AUTORES_COLUMN_INTEGER:
    assert(fieldType == AUTORES_FIELD_INTEGER);
    break;
  case AUTORES_COLUMN_INTEGER1D:
    assert(fieldType == AUTORES_FIELD_INTEGER1D);
    assert(AutoRes_Integer1DField_size 
	   ((AutoRes_Integer1DField *)field)
	   == AutoRes_Integer1DColumn_size 
	   ((AutoRes_Integer1DColumn *)column));
    break;
  case AUTORES_COLUMN_REAL:
    assert(fieldType == AUTORES_FIELD_REAL);
    break;
  case AUTORES_COLUMN_REAL1D:
    assert(fieldType == AUTORES_FIELD_REAL1D);
    assert(AutoRes_Real1DField_size 
	   ((AutoRes_Real1DField *)field)
	   == AutoRes_Real1DColumn_size 
	   ((AutoRes_Real1DColumn *)column));
    break;
  default:
    assert(0);
    break;
  }
  
  assert(self->nColumns < AUTORES_MAX_TABLE_IN_OUT_COLUMNS);
  assert(self->table == AutoRes_Column_table (column, columnType));
  
  self->columns[self->nColumns] = column;
  self->columnTypes[self->nColumns] = columnType;
  self->fields[self->nColumns] = field;
  self->fieldTypes[self->nColumns] = fieldType;
  self->nColumns++;
}

void AutoRes_TableInOut_associateColumnOfNameWithFieldOfName 
(AutoRes_TableInOut *self,
 const char *columnName,
 const char *fieldName)
{
  int columnId;
  void *column;
  AutoRes_ColumnType columnType;
  int fieldId;
  void *field;
  AutoRes_FieldType fieldType;
  
  assert(0 < strlen (columnName));
  assert(0 < strlen (fieldName));
  
  columnId = AutoRes_Table_columnIdOfName (self->table, columnName);
  assert(columnId != -1);
  column = AutoRes_Table_column (self->table, columnId);
  columnType = AutoRes_Table_columnType (self->table, columnId);
  
  fieldId = AutoRes_InOut_recordFieldIdOfName (self->inOut, fieldName);
  assert(fieldId != -1);
  field = AutoRes_InOut_recordField (self->inOut, fieldId);
  fieldType = AutoRes_InOut_recordFieldType (self->inOut, fieldId);

  AutoRes_TableInOut_associateColumnWithField 
    (self, column, columnType, field, fieldType);
}

void AutoRes_TableInOut_associateIntegerColumn 
(AutoRes_TableInOut *self,
 AutoRes_IntegerColumn *column)
{
  char columnName[AUTORES_MAX_COLUMN_NAME_LENGTH];
  int fieldId;
  AutoRes_IntegerField *field = NULL;

  AutoRes_IntegerColumn_getName 
    (column,
     columnName);
  fieldId = AutoRes_InOut_recordFieldIdOfName 
    (self->inOut, columnName);
  
  if (fieldId != -1) {
    assert(AutoRes_InOut_recordFieldType (self->inOut, fieldId)
	   == AUTORES_FIELD_INTEGER);
    field = (AutoRes_IntegerField *)AutoRes_InOut_recordField
      (self->inOut, fieldId);
  } else {
    field = AutoRes_IntegerField_New (columnName);
    AutoRes_InOut_addIntegerFieldIntoRecord (self->inOut, field);
  }
  
  assert(field != NULL);
  AutoRes_TableInOut_associateColumnWithField 
    (self, 
     column, AUTORES_COLUMN_INTEGER,
     field, AUTORES_FIELD_INTEGER);
}

void AutoRes_TableInOut_associateInteger1DColumn 
(AutoRes_TableInOut *self,
 AutoRes_Integer1DColumn *column)
{
  char columnName[AUTORES_MAX_COLUMN_NAME_LENGTH];
  int fieldId;
  AutoRes_Integer1DField *field = NULL;

  AutoRes_Integer1DColumn_getName 
    (column,
     columnName);
  fieldId = AutoRes_InOut_recordFieldIdOfName 
    (self->inOut, columnName);
  
  if (fieldId != -1) {
    assert(AutoRes_InOut_recordFieldType (self->inOut, fieldId)
	   == AUTORES_FIELD_INTEGER1D);
    field = (AutoRes_Integer1DField *)AutoRes_InOut_recordField
      (self->inOut, fieldId);
  } else {
    int size = AutoRes_Integer1DColumn_size (column);

    field = AutoRes_Integer1DField_New (columnName, size);
    AutoRes_InOut_addInteger1DFieldIntoRecord (self->inOut, field);
  }
  
  assert(field != NULL);
  AutoRes_TableInOut_associateColumnWithField 
    (self, 
     column, AUTORES_COLUMN_INTEGER1D,
     field, AUTORES_FIELD_INTEGER1D);
}

void AutoRes_TableInOut_associateRealColumn 
(AutoRes_TableInOut *self,
 AutoRes_RealColumn *column)
{
  char columnName[AUTORES_MAX_COLUMN_NAME_LENGTH];
  int fieldId;
  AutoRes_RealField *field = NULL;

  AutoRes_RealColumn_getName 
    (column,
     columnName);
  fieldId = AutoRes_InOut_recordFieldIdOfName 
    (self->inOut, columnName);
  
  if (fieldId != -1) {
    assert(AutoRes_InOut_recordFieldType (self->inOut, fieldId)
	   == AUTORES_FIELD_REAL);
    field = (AutoRes_RealField *)AutoRes_InOut_recordField
      (self->inOut, fieldId);
  } else {
    field = AutoRes_RealField_New (columnName);
    AutoRes_InOut_addRealFieldIntoRecord (self->inOut, field);
  }
  
  assert(field != NULL);
  AutoRes_TableInOut_associateColumnWithField 
    (self, 
     column, AUTORES_COLUMN_REAL,
     field, AUTORES_FIELD_REAL);
}

void AutoRes_TableInOut_associateReal1DColumn 
(AutoRes_TableInOut *self,
 AutoRes_Real1DColumn *column)
{
  char columnName[AUTORES_MAX_COLUMN_NAME_LENGTH];
  int fieldId;
  AutoRes_Real1DField *field = NULL;

  AutoRes_Real1DColumn_getName 
    (column,
     columnName);
  fieldId = AutoRes_InOut_recordFieldIdOfName 
    (self->inOut, columnName);
  
  if (fieldId != -1) {
    assert(AutoRes_InOut_recordFieldType (self->inOut, fieldId)
	   == AUTORES_FIELD_REAL1D);
    field = (AutoRes_Real1DField *)AutoRes_InOut_recordField
      (self->inOut, fieldId);
  } else {
    int size = AutoRes_Real1DColumn_size (column);

    field = AutoRes_Real1DField_New (columnName, size);
    AutoRes_InOut_addReal1DFieldIntoRecord (self->inOut, field);
  }
  
  assert(field != NULL);
  AutoRes_TableInOut_associateColumnWithField 
    (self, 
     column, AUTORES_COLUMN_REAL1D,
     field, AUTORES_FIELD_REAL1D);
}

int /* boolean */
AutoRes_TableInOut_open 
(AutoRes_TableInOut *self,
 const char *fileName,
 int writeFlag,
 int binaryFlag)
{
  int flag;
  
  flag = AutoRes_InOut_open (self->inOut, fileName, writeFlag, binaryFlag);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoRes_TableInOut_close 
(AutoRes_TableInOut *self)
{
  int flag;
  
  flag = AutoRes_InOut_close (self->inOut);
  if (!flag) return 0;

  return 1;
}

int /* boolean */
AutoRes_TableInOut_doHeaderSection 
(AutoRes_TableInOut *self)
{
  AutoRes_File *file;
  int writeFlag;
  int flag;
  
  file = AutoRes_InOut_file (self->inOut);
  assert(file != NULL);
  writeFlag = AutoRes_File_writeFlag (file);

  if (writeFlag) {
    int nRecords = AutoRes_Table_nRecords (self->table);
    
    AutoRes_InOut_setNRecords (self->inOut, nRecords);
  }

  flag = AutoRes_InOut_doHeaderSection (self->inOut);
  if (!flag) return 0;

  if (!writeFlag) {
    int nRecords = AutoRes_InOut_nRecords (self->inOut);

    if (AutoRes_Table_nRecords (self->table) == 0) {
      assert(AutoRes_Table_nReservedRecords (self->table) == 0);

      flag = AutoRes_Table_allocateRecords (self->table, nRecords);
      if (!flag) return 0;
    }
    assert(AutoRes_Table_nRecords (self->table) == nRecords);
  }

  return 1;
}

static void DoPreColumn 
(AutoRes_TableInOut *self,
 int recordId, 
 int columnId)
{
  void *column = self->columns[columnId];
  AutoRes_ColumnType columnType = self->columnTypes[columnId];
  void *field = self->fields[columnId];
  
  switch (columnType) {
  case AUTORES_COLUMN_INTEGER:
    {
      AutoRes_IntegerColumn *column2 = (AutoRes_IntegerColumn *)column;
      AutoRes_IntegerField *field2 = (AutoRes_IntegerField *)field;
      int value = AUTORES_I(column2, recordId);

      AutoRes_IntegerField_setValue (field2, value);
    }
    break;
  case AUTORES_COLUMN_INTEGER1D:
    {
      AutoRes_Integer1DColumn *column2 = (AutoRes_Integer1DColumn *)column;
      AutoRes_Integer1DField *field2 = (AutoRes_Integer1DField *)field;
      int size = AutoRes_Integer1DColumn_size (column2);
      int index;
      
      for (index = 0; index < size; index++) {
	int value = AUTORES_I1D(column2, recordId, index);
	
	AutoRes_Integer1DField_setValue (field2, index, value);
      }
    }
    break;
  case AUTORES_COLUMN_REAL:
    {
      AutoRes_RealColumn *column2 = (AutoRes_RealColumn *)column;
      AutoRes_RealField *field2 = (AutoRes_RealField *)field;
      double value = AUTORES_R(column2, recordId);

      AutoRes_RealField_setValue (field2, value);
    }
    break;
  case AUTORES_COLUMN_REAL1D:
    {
      AutoRes_Real1DColumn *column2 = (AutoRes_Real1DColumn *)column;
      AutoRes_Real1DField *field2 = (AutoRes_Real1DField *)field;
      int size = AutoRes_Real1DColumn_size (column2);
      int index;
      
      for (index = 0; index < size; index++) {
	double value = AUTORES_R1D(column2, recordId, index);
	
	AutoRes_Real1DField_setValue (field2, index, value);
      }
    }
    break;
  default:
    assert(0);
    break;
  }
}

static void DoPostColumn 
(AutoRes_TableInOut *self,
 int recordId, 
 int columnId)
{
  void *column = self->columns[columnId];
  AutoRes_ColumnType columnType = self->columnTypes[columnId];
  void *field = self->fields[columnId];
  
  switch (columnType) {
  case AUTORES_COLUMN_INTEGER:
    {
      AutoRes_IntegerColumn *column2 = (AutoRes_IntegerColumn *)column;
      AutoRes_IntegerField *field2 = (AutoRes_IntegerField *)field;
      int value = AutoRes_IntegerField_value (field2);

      AUTORES_I_SET(column2, recordId, value);
    }
    break;
  case AUTORES_COLUMN_INTEGER1D:
    {
      AutoRes_Integer1DColumn *column2 = (AutoRes_Integer1DColumn *)column;
      AutoRes_Integer1DField *field2 = (AutoRes_Integer1DField *)field;
      int size = AutoRes_Integer1DColumn_size (column2);
      int index;
      
      for (index = 0; index < size; index++) {
	int value = AutoRes_Integer1DField_value (field2, index);
	
	AUTORES_I1D_SET(column2, recordId, index, value);
      }
    }
    break;
  case AUTORES_COLUMN_REAL:
    {
      AutoRes_RealColumn *column2 = (AutoRes_RealColumn *)column;
      AutoRes_RealField *field2 = (AutoRes_RealField *)field;
      double value = AutoRes_RealField_value (field2);

      AUTORES_R_SET(column2, recordId, value);
    }
    break;
  case AUTORES_COLUMN_REAL1D:
    {
      AutoRes_Real1DColumn *column2 = (AutoRes_Real1DColumn *)column;
      AutoRes_Real1DField *field2 = (AutoRes_Real1DField *)field;
      int size = AutoRes_Real1DColumn_size (column2);
      int index;
      
      for (index = 0; index < size; index++) {
	double value = AutoRes_Real1DField_value (field2, index);
	
	AUTORES_R1D_SET(column2, recordId, index, value);
      }
    }
    break;
  default:
    assert(0);
    break;
  }
}

int /* boolean */
AutoRes_TableInOut_doRecordSection 
(AutoRes_TableInOut *self)
{
  AutoRes_File *file;
  int writeFlag;
  int nRecords;
  int iRecord;
  int iColumn;
  int flag;

  file = AutoRes_InOut_file (self->inOut);
  assert(file != NULL);
  writeFlag = AutoRes_File_writeFlag (file);

  nRecords = AutoRes_Table_nRecords (self->table);
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    if (writeFlag) {
      for (iColumn = 0; iColumn < self->nColumns; iColumn++) {
	DoPreColumn (self, iRecord, iColumn);
      }
    }
    
    flag = AutoRes_InOut_doEachOfRecordSection (self->inOut, iRecord);
    if (!flag) return 0;

    if (!writeFlag) {
      for (iColumn = 0; iColumn < self->nColumns; iColumn++) {
	DoPostColumn (self, iRecord, iColumn);
      }
    }
  }

  return 1;
}

int /* boolean */
AutoRes_TableInOut_do 
(AutoRes_TableInOut *self,
 const char *fileName,
 int writeFlag,
 int binaryFlag)
{
  int flag;

  flag = AutoRes_TableInOut_open (self, fileName, writeFlag, binaryFlag);
  if (!flag) return 0;
  
  flag = AutoRes_TableInOut_doHeaderSection (self);
  if (!flag) return 0;
  
  flag = AutoRes_TableInOut_doRecordSection (self);
  if (!flag) return 0;
  
  flag = AutoRes_TableInOut_close (self);
  if (!flag) return 0;
  
  return 1;
}

AutoRes_InOut *AutoRes_TableInOut_inOut 
(AutoRes_TableInOut *self)
{
  return self->inOut;
}



/* Table I/O Utility */

static AutoRes_TableInOut *CurrentTableInOut;
static int CurrentBinaryFlag = 0;
static int CurrentRecordIdFlag = 1;

void AutoRes_TableInOut_SetBinaryFlag (int flag)
{
  CurrentBinaryFlag = flag;
}

void AutoRes_TableInOut_SetRecordIdFlag (int flag)
{
  CurrentRecordIdFlag = flag;
}

void AutoRes_TableInOut_AssociateIntegerColumn 
(AutoRes_IntegerColumn *column)
{
  if (CurrentTableInOut == NULL) {
    AutoRes_Table *table = AutoRes_IntegerColumn_table (column);

    CurrentTableInOut = AutoRes_TableInOut_New (table);
  }
  assert(CurrentTableInOut != NULL);

  AutoRes_TableInOut_associateIntegerColumn 
    (CurrentTableInOut, column);
}

void AutoRes_TableInOut_AssociateInteger1DColumn 
(AutoRes_Integer1DColumn *column)
{
  if (CurrentTableInOut == NULL) {
    AutoRes_Table *table = AutoRes_Integer1DColumn_table (column);

    CurrentTableInOut = AutoRes_TableInOut_New (table);
  }
  assert(CurrentTableInOut != NULL);

  AutoRes_TableInOut_associateInteger1DColumn 
    (CurrentTableInOut, column);
}

void AutoRes_TableInOut_AssociateRealColumn 
(AutoRes_RealColumn *column)
{
  if (CurrentTableInOut == NULL) {
    AutoRes_Table *table = AutoRes_RealColumn_table (column);

    CurrentTableInOut = AutoRes_TableInOut_New (table);
  }
  assert(CurrentTableInOut != NULL);

  AutoRes_TableInOut_associateRealColumn 
    (CurrentTableInOut, column);
}

void AutoRes_TableInOut_AssociateReal1DColumn 
(AutoRes_Real1DColumn *column)
{
  if (CurrentTableInOut == NULL) {
    AutoRes_Table *table = AutoRes_Real1DColumn_table (column);

    CurrentTableInOut = AutoRes_TableInOut_New (table);
  }
  assert(CurrentTableInOut != NULL);

  AutoRes_TableInOut_associateReal1DColumn 
    (CurrentTableInOut, column);
}

int /* boolean */
AutoRes_TableInOut_Read (const char *fileName)
{
  AutoRes_InOut *inOut;
  int writeFlag = 0;
  int flag;

  assert(CurrentTableInOut != NULL);

  inOut = AutoRes_TableInOut_inOut (CurrentTableInOut);

  AutoRes_InOut_setRecordIdFlag (inOut, CurrentRecordIdFlag);

  flag = AutoRes_TableInOut_do 
    (CurrentTableInOut, fileName, writeFlag, CurrentBinaryFlag);
  
  AutoRes_TableInOut_delete (CurrentTableInOut);
  CurrentTableInOut = NULL;

  assert(CurrentTableInOut == NULL);

  return flag;
}

int /* boolean */
AutoRes_TableInOut_Write (const char *fileName)
{
  AutoRes_InOut *inOut;
  int writeFlag = 1;
  int flag;

  assert(CurrentTableInOut != NULL);

  inOut = AutoRes_TableInOut_inOut (CurrentTableInOut);

  AutoRes_InOut_setRecordIdFlag (inOut, CurrentRecordIdFlag);

  flag = AutoRes_TableInOut_do 
    (CurrentTableInOut, fileName, writeFlag, CurrentBinaryFlag);

  AutoRes_TableInOut_delete (CurrentTableInOut);
  CurrentTableInOut = NULL;

  assert(CurrentTableInOut == NULL);

  return flag;
}

int /* boolean */
AutoRes_TableInOut_WriteMetaData (const char *fileName)
{
  AutoRes_InOut *inOut;
  int flag;

  assert(CurrentTableInOut != NULL);

  inOut = AutoRes_TableInOut_inOut (CurrentTableInOut);

  flag = AutoRes_InOut_writeMetaData (inOut, fileName);

  AutoRes_TableInOut_delete (CurrentTableInOut);
  CurrentTableInOut = NULL;

  assert(CurrentTableInOut == NULL);

  return flag;
}



