/* autores_use_flag.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "autores_base.h"
#include "autores_memory.h"
#include "autores_column.h"
#include "autores_integer_column.h"
#include "autores_real_column.h"
#include "autores_use_flag.h"



/* Use Flag Collection */

#define COLUMN_NAME_USE_FLAG "AutoRes:useFlag"

AUTORES_IMPLEMENT_HANDLE(AutoRes_UseFlagCollection);

AutoRes_UseFlagCollection *AutoRes_UseFlagCollection_New 
(AutoRes_Table *table)
{
  AutoRes_UseFlagCollection *collection;

  collection = AutoRes_UseFlagCollection_NewImplementation ();

  collection->table = table;
  collection->useFlagColumn = NULL;
  
  return collection;
}

void AutoRes_UseFlagCollection_delete 
(AutoRes_UseFlagCollection *self)
{
  if (self->useFlagColumn != NULL) {
    AutoRes_Table_removeColumn 
      (self->table, 
       AutoRes_Table_columnIdOfColumn 
       (self->table, self->useFlagColumn));
    self->useFlagColumn = NULL;
  }

  AutoRes_ReleaseHandle (self->handle);
  self->handle = AUTORES_NULL_HANDLE;
}

int /* boolean */
AutoRes_UseFlagCollection_makeColumn 
(AutoRes_UseFlagCollection *self)
{
  int flag;

  assert(self->useFlagColumn == NULL);

  self->useFlagColumn = AutoRes_IntegerColumn_New 
    (self->table, COLUMN_NAME_USE_FLAG);
  assert(self->useFlagColumn != NULL);

  flag = AutoRes_Table_addIntegerColumn 
    (self->table, self->useFlagColumn);
  if (!flag) return 0;

  assert(self->useFlagColumn != NULL);

  return 1;
}

void AutoRes_UseFlagCollection_initializeColumn 
(AutoRes_UseFlagCollection *self)
{
  int nRecords = AutoRes_Table_nRecords (self->table);
  int iRecord;

  assert(self->useFlagColumn != NULL);

  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    AUTORES_I_SET(self->useFlagColumn, iRecord, 1);
  }
}

int AutoRes_UseFlagCollection_addRecord 
(AutoRes_UseFlagCollection *self)
{
  int newRecordId;

  assert(0 < AutoRes_Table_nReservedRecords (self->table));
  assert(self->useFlagColumn != NULL);

  newRecordId = AutoRes_Table_addRecord (self->table);
  if (newRecordId == -1) return -1;

  AUTORES_I_SET(self->useFlagColumn, newRecordId, 1);

  return newRecordId;
}

void AutoRes_UseFlagCollection_removeRecord 
(AutoRes_UseFlagCollection *self,
 int recordId)
{
  assert(self->useFlagColumn != NULL);
  assert(AUTORES_I(self->useFlagColumn, recordId) == 1);

  AUTORES_I_SET(self->useFlagColumn, recordId, 0);
}

static void PackIntegerColumn 
(AutoRes_Table *table,
 AutoRes_IntegerColumn *column,
 AutoRes_IntegerColumn *useFlagColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int iRecord;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int useFlag = AUTORES_I(useFlagColumn, iRecord);
    
    if (useFlag) {
      int value = AUTORES_I(column, iRecord);

      AUTORES_I_SET(column, count, value);
      count++;
    }
  }
}

static void PackInteger1DColumn 
(AutoRes_Table *table,
 AutoRes_Integer1DColumn *column,
 AutoRes_IntegerColumn *useFlagColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int size = AutoRes_Integer1DColumn_size (column);
  int iRecord;
  int index;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int useFlag = AUTORES_I(useFlagColumn, iRecord);
    
    if (useFlag) {
      for (index = 0; index < size; index++) {
	int value = AUTORES_I1D(column, iRecord, index);

	AUTORES_I1D_SET(column, count, index, value);
      }
      count++;
    }
  }
}

static void PackRealColumn 
(AutoRes_Table *table,
 AutoRes_RealColumn *column,
 AutoRes_IntegerColumn *useFlagColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int iRecord;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int useFlag = AUTORES_I(useFlagColumn, iRecord);
    
    if (useFlag) {
      double value = AUTORES_R(column, iRecord);

      AUTORES_R_SET(column, count, value);
      count++;
    }
  }
}

static void PackReal1DColumn 
(AutoRes_Table *table,
 AutoRes_Real1DColumn *column,
 AutoRes_IntegerColumn *useFlagColumn)
{
  int nRecords = AutoRes_Table_nRecords (table);
  int size = AutoRes_Real1DColumn_size (column);
  int iRecord;
  int index;
  int count;

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int useFlag = AUTORES_I(useFlagColumn, iRecord);
    
    if (useFlag) {
      for (index = 0; index < size; index++) {
	double value = AUTORES_R1D(column, iRecord, index);

	AUTORES_R1D_SET(column, count, index, value);
      }
      count++;
    }
  }
}

void AutoRes_UseFlagCollection_pack 
(AutoRes_UseFlagCollection *self)
{
  int nRecords;
  int nColumns;
  int iColumn;
  int iRecord;
  int count;

  assert(self->useFlagColumn != NULL);

  nRecords = AutoRes_Table_nRecords (self->table);
  nColumns = AutoRes_Table_nColumns (self->table);

  for (iColumn = 0; iColumn < nColumns; iColumn++) {
    void *column = AutoRes_Table_column (self->table, iColumn);
    AutoRes_ColumnType columnType 
      = AutoRes_Table_columnType (self->table, iColumn);

    if (column == (void *)self->useFlagColumn) {
      continue;
    }

    switch (columnType) {
    case AUTORES_COLUMN_INTEGER:
      PackIntegerColumn 
	(self->table, (AutoRes_IntegerColumn *)column, 
	 self->useFlagColumn);
      break;
    case AUTORES_COLUMN_INTEGER1D:
      PackInteger1DColumn 
	(self->table, (AutoRes_Integer1DColumn *)column, 
	 self->useFlagColumn);
      break;
    case AUTORES_COLUMN_REAL:
      PackRealColumn 
	(self->table, (AutoRes_RealColumn *)column, 
	 self->useFlagColumn);
      break;
    case AUTORES_COLUMN_REAL1D:
      PackReal1DColumn 
	(self->table, (AutoRes_Real1DColumn *)column, 
	 self->useFlagColumn);
      break;
    default:
      assert(0);
      break;
    }
  }

  count = 0;
  for (iRecord = 0; iRecord < nRecords; iRecord++) {
    int useFlag = AUTORES_I(self->useFlagColumn, iRecord);

    if (useFlag) {
      count++;
    }
  }
  for (iRecord = 0; iRecord < count; iRecord++) {
    AUTORES_I_SET(self->useFlagColumn, iRecord, 1);
  }
  for (iRecord = count; iRecord < nRecords; iRecord++) {
    AUTORES_I_SET(self->useFlagColumn, iRecord, 0);
  }
  AutoRes_Table_packNRecords (self->table, count);
}

AutoRes_Table *AutoRes_UseFlagCollection_table 
(AutoRes_UseFlagCollection *self)
{
  return self->table;
}

int /* boolean */
AutoRes_UseFlagCollection_recordIsUsed 
(AutoRes_UseFlagCollection *self,
 int recordId)
{
  return AUTORES_I(self->useFlagColumn, recordId);
}



